__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import copy

import pandas

import numpy as np

import datetime
from datetime import timedelta

import pandas as pd

from findatapy.market.ioengine import IOEngine
from findatapy.market.marketdatarequest import MarketDataRequest
from findatapy.timeseries import Filter, Calculations
from findatapy.util import DataConstants, LoggerManager, ConfigManager, SwimPool

constants = DataConstants()

class MarketDataGenerator(object):
    """Returns market data time series by directly calling market data sources.

    At present it supports Bloomberg (bloomberg), Yahoo (yahoo), Quandl (quandl), FRED (fred) etc. which are implemented
    in subclasses of DataVendor class. This provides a common wrapper for all these data sources.
    """

    def __init__(self):
        self.config = ConfigManager().get_instance()
        self.filter = Filter()
        self.calculations = Calculations()
        self.io_engine = IOEngine()
        self._intraday_code = -1
        self.days_expired_intraday_contract_download = -1

        return

    def set_intraday_code(self, code):
        self._intraday_code = code

    def get_data_vendor(self, md_request):
        """Loads appropriate data service class

        Parameters
        ----------
        md_request : MarketDataRequest
            the data_source to use "bloomberg", "quandl", "yahoo", "google", "fred" etc.
            we can also have forms like "bloomberg-boe" separated by hyphens

        Returns
        -------
        DataVendor
        """
        logger = LoggerManager().getLogger(__name__)

        data_source = md_request.data_source
        data_engine = md_request.data_engine

        # Special case for files (csv, h5, parquet or zip)
        if '.csv' in str(data_source) or '.h5' in str(data_source) or '.parquet' in str(data_source) or '.zip' in str(data_source) \
                or data_engine is not None:
            from findatapy.market.datavendorweb import DataVendorFlatFile
            data_vendor = DataVendorFlatFile()
        else:
            try:
                data_source = data_source.split("-")[0]
            except:
                logger.error("Was data data_source specified?")

                return None

            if data_source == 'bloomberg':
                try:
                    from findatapy.market.datavendorbbg import DataVendorBBGOpen
                    data_vendor = DataVendorBBGOpen()
                except:
                    logger.warn("Bloomberg needs to be installed")

            elif data_source == 'quandl':
                from findatapy.market.datavendorweb import DataVendorQuandl
                data_vendor = DataVendorQuandl()

            elif data_source == 'eikon':
                from findatapy.market.datavendorweb import DataVendorEikon
                data_vendor = DataVendorEikon()

            elif data_source == 'ons':
                from findatapy.market.datavendorweb import DataVendorONS
                data_vendor = DataVendorONS()

            elif data_source == 'boe':
                from findatapy.market.datavendorweb import DataVendorBOE
                data_vendor = DataVendorBOE()

            elif data_source == 'dukascopy':
                from findatapy.market.datavendorweb  import DataVendorDukasCopy
                data_vendor = DataVendorDukasCopy()

            elif data_source == 'fxcm':
                from findatapy.market.datavendorweb  import DataVendorFXCM
                data_vendor = DataVendorFXCM()

            elif data_source == 'alfred':
                from findatapy.market.datavendorweb  import DataVendorALFRED
                data_vendor = DataVendorALFRED()

            elif data_source == 'yahoo':
                from findatapy.market.datavendorweb  import DataVendorYahoo
                data_vendor = DataVendorYahoo()

            elif data_source in ['google', 'fred', 'oecd', 'eurostat', 'edgar-index']:
                from findatapy.market.datavendorweb  import DataVendorPandasWeb
                data_vendor = DataVendorPandasWeb()

            elif data_source == 'bitcoincharts':
                from findatapy.market.datavendorweb import DataVendorBitcoincharts
                data_vendor = DataVendorBitcoincharts()
            elif data_source == 'poloniex':
                from findatapy.market.datavendorweb import DataVendorPoloniex
                data_vendor = DataVendorPoloniex()
            elif data_source == 'binance':
                from findatapy.market.datavendorweb import DataVendorBinance
                data_vendor = DataVendorBinance()
            elif data_source == 'bitfinex':
                from findatapy.market.datavendorweb import DataVendorBitfinex
                data_vendor = DataVendorBitfinex()
            elif data_source == 'gdax':
                from findatapy.market.datavendorweb import DataVendorGdax
                data_vendor = DataVendorGdax()
            elif data_source == 'kraken':
                from findatapy.market.datavendorweb import DataVendorKraken
                data_vendor = DataVendorKraken()
            elif data_source == 'bitmex':
                from findatapy.market.datavendorweb import DataVendorBitmex
                data_vendor = DataVendorBitmex()
            elif data_source == 'alphavantage':
                from findatapy.market.datavendorweb import DataVendorAlphaVantage
                data_vendor = DataVendorAlphaVantage()
            elif data_source == 'huobi':
                from findatapy.market.datavendorweb import DataVendorHuobi
                data_vendor = DataVendorHuobi()

        # TODO add support for other data sources (like Reuters)

        return data_vendor

    def fetch_market_data(self, market_data_request, kill_session = True):
        """Loads time series from specified data provider

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains various properties describing time series to fetched, including ticker, start & finish date etc.

        Returns
        -------
        pandas.DataFrame
        """
        logger = LoggerManager().getLogger(__name__)

        # data_vendor = self.get_data_vendor(md_request.data_source)

        # check if tickers have been specified (if not load all of them for a category)
        # also handle single tickers/list tickers
        create_tickers = False

        if market_data_request.vendor_tickers is not None and market_data_request.tickers is None:
            market_data_request.tickers = market_data_request.vendor_tickers

        tickers = market_data_request.tickers

        if tickers is None:
            create_tickers = True
        elif isinstance(tickers, str):
            if tickers == '': create_tickers = True
        elif isinstance(tickers, list):
            if tickers == []: create_tickers = True

        if create_tickers:
            market_data_request.tickers = ConfigManager().get_instance().get_tickers_list_for_category(
            market_data_request.category, market_data_request.data_source, market_data_request.freq, market_data_request.cut)

        # intraday or tick: only one ticker per cache file
        if (market_data_request.freq in ['intraday', 'tick', 'second', 'hour', 'minute']):
            data_frame_agg = self.download_intraday_tick(market_data_request)
     #       return data_frame_agg

        # Daily: multiple tickers per cache file - assume we make one API call to vendor library
        else:
            data_frame_agg = self.download_daily(market_data_request)

        if('internet_load' in market_data_request.cache_algo):
            logger.debug("Internet loading.. ")

            # Signal to data_vendor template to exit session
            # if data_vendor is not None and kill_session == True: data_vendor.kill_session()

        if(market_data_request.cache_algo == 'cache_algo'):
            logger.debug("Only caching data in memory, do not return any time series."); return

        # Only return time series if specified in the algo
        if 'return' in market_data_request.cache_algo:
            # Special case for events/events-dt which is not indexed like other tables (also same for downloading futures
            # contracts dates)
            if market_data_request.category is not None:
                if 'events' in market_data_request.category:
                    return data_frame_agg

            # Pad columns a second time (is this necessary to do here again?)
            # TODO only do this for not daily data?
            try:
                if data_frame_agg is not None:
                    data_frame_agg = self.filter.filter_time_series(market_data_request, data_frame_agg, pad_columns=True)
                    data_frame_agg = data_frame_agg.dropna(how='all')

                    # Resample data using pandas if specified in the MarketDataRequest
                    if market_data_request.resample is not None:
                        if 'last' in market_data_request.resample_how:
                            data_frame_agg = data_frame_agg.resample(market_data_request.resample).last()
                        elif 'first' in market_data_request.resample_how:
                            data_frame_agg = data_frame_agg.resample(market_data_request.resample).first()

                        if 'dropna' in market_data_request.resample_how:
                            data_frame_agg = data_frame_agg.dropna(how='all')
                else:
                    logger.warn("No data returned for " + str(market_data_request.tickers))

                return data_frame_agg
            except Exception as e:
                print(str(e))
                if data_frame_agg is not None:
                    return data_frame_agg

                import traceback

                logger.warn("No data returned for " + str(market_data_request.tickers))

                return None

    def create_time_series_hash_key(self, market_data_request, ticker = None):
        """Creates a hash key for retrieving the time series

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains various properties describing time series to fetched, including ticker, start & finish date etc.

        Returns
        -------
        str
        """

        if(isinstance(ticker, list)):
            ticker = ticker[0]

        return self.create_cache_file_name(MarketDataRequest().create_category_key(
            md_request=market_data_request, ticker=ticker))

    def download_intraday_tick(self, market_data_request):
        """Loads intraday time series from specified data provider

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains various properties describing time series to fetched, including ticker, start & finish date etc.

        Returns
        -------
        pandas.DataFrame
        """

        data_frame_agg = None
        calcuations = Calculations()

        ticker_cycle = 0

        data_frame_group = []

        # Single threaded version
        # handle intraday ticker calls separately one by one
        if len(market_data_request.tickers) == 1 or constants.market_thread_no['other'] == 1:
            for ticker in market_data_request.tickers:
                market_data_request_single = copy.copy(market_data_request)
                market_data_request_single.tickers = ticker

                if market_data_request.vendor_tickers is not None:
                    market_data_request_single.vendor_tickers = [market_data_request.vendor_tickers[ticker_cycle]]
                    ticker_cycle = ticker_cycle + 1

                # We downscale into float32, to avoid memory problems in Python (32 bit)
                # data is stored on disk as float32 anyway
                # old_finish_date = market_data_request_single.finish_date
                #
                # market_data_request_single.finish_date = self.refine_expiry_date(md_request)
                #
                # if market_data_request_single.finish_date >= market_data_request_single.start_date:
                #     data_frame_single = data_vendor.load_ticker(market_data_request_single)
                # else:
                #     data_frame_single = None
                #
                # market_data_request_single.finish_date = old_finish_date
                #
                # data_frame_single = data_vendor.load_ticker(market_data_request_single)

                data_frame_single = self.fetch_single_time_series(market_data_request)

                # If the vendor doesn't provide any data, don't attempt to append
                if data_frame_single is not None:
                    if data_frame_single.empty == False:
                        data_frame_single.index.name = 'Date'
                        data_frame_single = data_frame_single.astype('float32')

                        data_frame_group.append(data_frame_single)

                        # # if you call for returning multiple tickers, be careful with memory considerations!
                        # if data_frame_agg is not None:
                        #     data_frame_agg = data_frame_agg.join(data_frame_single, how='outer')
                        # else:
                        #     data_frame_agg = data_frame_single

                # key = self.create_category_key(md_request, ticker)
                # fname = self.create_cache_file_name(key)
                # self._time_series_cache[fname] = data_frame_agg  # cache in memory (disable for intraday)

            # If you call for returning multiple tickers, be careful with memory considerations!
            if data_frame_group is not None:
                data_frame_agg = calcuations.join(data_frame_group, how='outer')

            return data_frame_agg

        else:
            market_data_request_list = []

            # Create a list of MarketDataRequests
            for ticker in market_data_request.tickers:
                market_data_request_single = copy.copy(market_data_request)
                market_data_request_single.tickers = ticker

                if market_data_request.vendor_tickers is not None:
                    market_data_request_single.vendor_tickers = [market_data_request.vendor_tickers[ticker_cycle]]
                    ticker_cycle = ticker_cycle + 1

                market_data_request_list.append(market_data_request_single)

            return self.fetch_group_time_series(market_data_request_list)

    def fetch_single_time_series(self, market_data_request):
        logger = LoggerManager().getLogger(__name__)
        market_data_request = MarketDataRequest(md_request=market_data_request)

        # Only includes those tickers have not expired yet!
        start_date = pd.Timestamp(market_data_request.start_date).date()

        current_date = pd.Timestamp(datetime.datetime.utcnow().date())

        tickers = market_data_request.tickers
        vendor_tickers = market_data_request.vendor_tickers

        expiry_date = pd.Timestamp(market_data_request.expiry_date)

        config = ConfigManager().get_instance()

        # In many cases no expiry is defined so skip them
        for i in range(0, len(tickers)):
            try:
                expiry_date = config.get_expiry_for_ticker(market_data_request.data_source, tickers[i])
            except:
                pass

            if expiry_date is not None:
                expiry_date = pd.Timestamp(expiry_date)

                if not(pd.isna(expiry_date)):
                    # Use pandas Timestamp, a bit more robust with weird dates (can fail if comparing date vs datetime)
                    # if the expiry is before the start date of our download don't bother downloading this ticker
                    if expiry_date < start_date:
                        tickers[i] = None

                    # Special case for futures-contracts which are intraday
                    # avoid downloading if the expiry date is very far in the past
                    # (we need this before there might be odd situations where we run on an expiry date, but still want to get
                    # data right till expiry time)
                    if market_data_request.category == 'futures-contracts' and market_data_request.freq == 'intraday' \
                            and self.days_expired_intraday_contract_download > 0:

                        if expiry_date + pd.Timedelta(days=self.days_expired_intraday_contract_download) < current_date:
                            tickers[i] = None

                    if vendor_tickers is not None and tickers[i] is None:
                        vendor_tickers[i] = None

        market_data_request.tickers = [e for e in tickers if e != None]

        if vendor_tickers is not None:
            market_data_request.vendor_tickers = [e for e in vendor_tickers if e != None]

        data_frame_single = None

        if len(market_data_request.tickers) > 0:
            data_frame_single = self.get_data_vendor(market_data_request).load_ticker(market_data_request)
            #print(data_frame_single.head(n=10))

        if data_frame_single is not None:
            if data_frame_single.empty == False:
                data_frame_single.index.name = 'Date'

                # Will fail for DataFrames which includes dates/strings (eg. futures contract names)
                data_frame_single = Calculations().convert_to_numeric_dataframe(data_frame_single)

                if market_data_request.freq == "second":
                    data_frame_single = data_frame_single.resample("1s")

        return data_frame_single

    def fetch_group_time_series(self, market_data_request_list):

        logger = LoggerManager().getLogger(__name__)

        data_frame_agg = None

        thread_no = constants.market_thread_no['other']

        if market_data_request_list[0].data_source in constants.market_thread_no:
            thread_no = constants.market_thread_no[market_data_request_list[0].data_source]

        if thread_no > 0:
            pool = SwimPool().create_pool(thread_technique = constants.market_thread_technique, thread_no=thread_no)

            # Open the market data downloads in their own threads and return the results
            result = pool.map_async(self.fetch_single_time_series, market_data_request_list)
            data_frame_group = result.get()

            pool.close()
            pool.join()
        else:
            data_frame_group = []

            for md_request in market_data_request_list:
                data_frame_group.append(self.fetch_single_time_series(md_request))

        # Collect together all the time series
        if data_frame_group is not None:
            data_frame_group = [i for i in data_frame_group if i is not None]

            # import itertools
            # columns = list(itertools.chain.from_iterable([i.columns for i in data_frame_group if i is not None]))

            # For debugging!
            # import pickle
            # import datetime
            # pickle.dump(data_frame_group, open(str(datetime.datetime.now()).replace(':', '-').replace(' ', '-').replace(".", "-") + ".p", "wb"))

            if data_frame_group is not None:
                try:
                    data_frame_agg = self.calculations.join(data_frame_group, how='outer')

                    # Force ordering to be the same!
                    # data_frame_agg = data_frame_agg[columns]
                except Exception as e:
                    logger.warning('Possible overlap of columns? Have you specifed same ticker several times: ' + str(e))

        return data_frame_agg

    def download_daily(self, market_data_request):
        """Loads daily time series from specified data provider

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains various properties describing time series to fetched, including ticker, start & finish date etc.

        Returns
        -------
        pandas.DataFrame
        """

        key = MarketDataRequest().create_category_key(md_request=market_data_request)

        is_key_overriden = False

        for k in constants.override_multi_threading_for_categories:
            if k in key:
                is_key_overriden = True
                break

        # By default use other
        thread_no = constants.market_thread_no['other']

        if str(market_data_request.data_source) in constants.market_thread_no:
            thread_no = constants.market_thread_no[market_data_request.data_source]

        # Daily data does not include ticker in the key, as multiple tickers in the same file
        if thread_no == 1 or '.csv' in str(market_data_request.data_source) or \
            '.h5' in str(market_data_request.data_source) or '.parquet' in str(market_data_request.data_source) \
                or '.zip' in str(market_data_request.data_source) or market_data_request.data_engine is not None:
            # data_frame_agg = data_vendor.load_ticker(md_request)
            data_frame_agg = self.fetch_single_time_series(market_data_request)
        else:
            market_data_request_list = []
            
            # When trying your example 'equitiesdata_example' I had a -1 result so it went out of the comming loop and I had errors in execution
            group_size = max(int(len(market_data_request.tickers) / thread_no - 1),0)

            if group_size == 0: group_size = 1

            # Split up tickers into groups related to number of threads to call
            for i in range(0, len(market_data_request.tickers), group_size):
                market_data_request_single = copy.copy(market_data_request)
                market_data_request_single.tickers = market_data_request.tickers[i:i + group_size]

                if market_data_request.vendor_tickers is not None:
                    market_data_request_single.vendor_tickers = \
                        market_data_request.vendor_tickers[i:i + group_size]

                market_data_request_list.append(market_data_request_single)

            # Special case where we make smaller calls one after the other
            if is_key_overriden:

                data_frame_list = []

                for md in market_data_request_list:
                    data_frame_list.append(self.fetch_single_time_series(md))

                data_frame_agg = self.calculations.join(data_frame_list, how='outer')
            else:
                data_frame_agg = self.fetch_group_time_series(market_data_request_list)

        # fname = self.create_cache_file_name(key)
        # self._time_series_cache[fname] = data_frame_agg  # cache in memory (ok for daily data)

        return data_frame_agg

    def refine_expiry_date(self, market_data_request):

        # Expiry date
        if market_data_request.expiry_date is None:
            ConfigManager().get_instance().get_expiry_for_ticker(market_data_request.data_source, market_data_request.ticker)

        return market_data_request

    def create_cache_file_name(self, filename):
        return constants.folder_time_series_data + "/" + filename



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

from findatapy.util.singleton import Singleton

class CacheManager(object):
    __metaclass__ = Singleton

    _is_init = 0

    _dict_cache = {}

    def __init__(self, *args, **kwargs):
        if CacheManager._is_init == 0:
            # CacheManager.populate_time_series_dictionaries()
            CacheManager._is_init = 1
        pass

    ### time series ticker manipulators
    @staticmethod
    def add_cache(key, obj):
        CacheManager._dict_cache[key] = obj

    @staticmethod
    def get_cache(key):
        if key in CacheManager._dict_cache:
            # print('cached ' + key)
            return CacheManager._dict_cache[key]

        return None

    @staticmethod
    def is_in_cache(key):
        return key in CacheManager._dict_cache

    @staticmethod
    def flush_cache():
        CacheManager._dict_cache = {}


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download free tick data from DukasCopy example
    # run_example = 2 - download free FX daily data from Quandl
    # run_example = 3 - download FX data from FRED
    # run_example = 4 - download FX data from Bloomberg
    # run_example = 5 - download second FX data from Bloomberg
    # run_example = 6 - download free tick data from FXCM example (compare with DukasCopy)

    run_example = 4

    if run_example == 1 or run_example == 0:

        ####### DukasCopy examples
        # let's download data for 14 Jun 2016 for EUR/USD - the raw data has bid/ask, if we specify close, we calculate
        # it as the average

        # first we can do it by defining all the vendor fields, tickers etc. so we bypass the configuration file
        md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='15 Jun 2016',
                                       fields=['bid'], vendor_fields=['bid'],
                                       freq='tick', data_source='dukascopy',
                                       tickers=['EURUSD'], vendor_tickers=['EURUSD'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # now let's do it using the category keyword, which goes into our config files (only works for predefined tickers!)
        # simplifies our calling procedure a lot!
        md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='15 Jun 2016',
                                       category='fx', fields=['close'], freq='tick', data_source='dukascopy', tickers=['EURUSD'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # we can also get the bid/ask quotes directly
        md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='15 Jun 2016',
                                       category='fx', fields=['bid', 'ask'], freq='tick', data_source='dukascopy', tickers=['EURUSD'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))


    if run_example == 2 or run_example == 0:
        ####### Quandl data examples

        # for this to work make sure you edit the Quandl API key in DataConstants file

        # already defined for us are the tickers for G10 USD FX in various CSV files in conf folder
        # by default, assume data is for 'close' field only, cut as being 'NYC' etc.
        # we can use keywords, 'month' and 'year' to specify the last year of data, alternatively, we can put a specific date
        md_request = MarketDataRequest(start_date='year', category='fx', data_source='quandl', tickers=['EURUSD'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # if we give it a cross which doesn't exist in the database directly, it will try to synthesis from the USD rates
        # this might be ok for daily data, but should not be used for tick data
        # eg. AUD/JPY
        md_request = MarketDataRequest(start_date='year', category='fx', data_source='quandl', tickers=['AUDJPY'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # if you don't specify tickers, it will download all the tickers in that data_source.freq.category.cut combination
        # ie. quandl.daily.fx.NYC (all G10 USD crosses)
        md_request = MarketDataRequest(start_date='month', category='fx', data_source='quandl', cut='NYC')

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

    if run_example == 3 or run_example == 0:
        ####### FRED example

        # if we give it a cross which doesn't exist in the database directly, it will try to synthesis from the USD rates
        # this might be ok for daily data, but should not be used for tick data
        # eg. AUD/JPY
        md_request = MarketDataRequest(start_date='year', category='fx', data_source='quandl', tickers=['AUDJPY'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # if you don't specify tickers, it will download all the tickers in that data_source.freq.category.cut combination
        # ie. quandl.daily.fx.NYC (all G10 USD crosses)
        md_request = MarketDataRequest(start_date='month', category='fx', data_source='quandl', cut='NYC')

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

    if run_example == 4 or run_example == 0:
        ####### Bloomberg

        # now we will try various examples with Bloomberg
        # only works if you have Bloomberg terminal installed and the Python API!
        md_request = MarketDataRequest(start_date='week', category='fx', data_source='bloomberg', tickers=['AUDJPY'])
        df = market.fetch_market(md_request)

        print(df.tail(n=10))

        # let's now try downloading 1 minute intraday data for the past week from Bloomberg
        md_request = MarketDataRequest(start_date='week', freq='intraday', category='fx', data_source='bloomberg', tickers=['AUDJPY'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # let's now try downloading 1 minute intraday data for the past week from Bloomberg, directly specifying tickers
        # bypassing our config file
        # this is handy when you want to download something which isn't predefined in the configuration files
        md_request = MarketDataRequest(start_date='week', freq='intraday', data_source='bloomberg',
                                       tickers=['AUDJPY'], vendor_tickers=['AUDJPY BGN Curncy'],
                                       fields=['close'], vendor_fields=['close'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))


    if run_example == 5 or run_example == 0:
        ####### Bloomberg

        # let's now try downloading tick data for the past hour from Bloomberg, directly specifying tickers
        # bypassing our config file
        # this is handy when you want to download something which isn't predefined in the configuration files
        # then convert into seconds data using pandas
        md_request = MarketDataRequest(start_date='hour', freq='tick', data_source='bloomberg',
                                       tickers=['AUDJPY'], vendor_tickers=['AUDJPY BGN Curncy'],
                                       fields=['close'], vendor_fields=['close'])

        df = market.fetch_market(md_request)
        df = df.resample('1s').mean()

        print(df.tail(n=60))

    if run_example == 6 or run_example == 0:
        ####### FXCM (and compare with DukasCopy) examples

        # let's download data for end of 2016/start 2017 for EUR/USD - the raw data has bid/ask, if we specify close, we calculate
        # it as the average

        # first we can do it by defining all the vendor fields, tickers etc. so we bypass the configuration file
        md_request = MarketDataRequest(start_date='01 Dec 2016', finish_date='07 Dec 2016',
                                       fields=['bid'], vendor_fields=['bid'],
                                       freq='tick', data_source='fxcm',
                                       tickers=['EURUSD'], vendor_tickers=['EURUSD'])

        df_tick = market.fetch_market(md_request)
        df = df_tick.resample('1min').mean()

        md_request.data_source = 'dukascopy'
        df1 = market.fetch_market(md_request).resample('1min').mean()

        df1.columns = [c + '_dk' for c in df1.columns]
        df = df1.join(df)

        print(df_tick.tail(n=100))
        print(df.tail(n=100))


__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import numpy
import pytz
import pandas.tseries.offsets

class Timezone(object):
    """Various functions wrapping onto pandas and pytz for quickly converting timezones for dataframes.

    """

    def convert_index_from_UTC_to_new_york_time(self, data_frame):
        new_york = pytz.timezone('America/New_York')

        try:
            data_frame = data_frame.tz_localize(pytz.utc)
        except:
            pass

        data_frame = data_frame.tz_convert(new_york)

        return data_frame

    def convert_index_from_UTC_to_london_time(self, data_frame):
        london = pytz.timezone('Europe/London')

        try:
            data_frame = data_frame.tz_localize(pytz.utc)
        except:
            pass

        data_frame = data_frame.tz_convert(london)

        return data_frame

    def convert_index_time_zone(self, data_frame, from_tz, to_tz):
        data_frame = data_frame.tz_localize(pytz.timezone(from_tz))\
            .tz_convert(pytz.timezone(to_tz))

        return data_frame

    def convert_index_from_UTC_to_alt(self, data_frame, time_zone):
        alt = pytz.timezone(time_zone)
        data_frame = data_frame.tz_localize(pytz.utc).tz_convert(alt)

        return data_frame

    def convert_index_aware_to_UTC_time(self, data_frame):
        utc = pytz.timezone('UTC')
        data_frame = data_frame.tz_convert(utc)

        return data_frame

    def convert_index_aware_to_new_york_time(self, data_frame):
        new_york = pytz.timezone('America/New_York')
        data_frame = data_frame.tz_convert(new_york)

        return data_frame

    def convert_index_aware_to_london_time(self, data_frame):
        london = pytz.timezone('Europe/London')
        data_frame = data_frame.tz_convert(london)

        return data_frame

    def convert_index_aware_to_alt(self, data_frame, time_zone):
        try:
            alt = pytz.timezone(time_zone)
        except:
            alt = time_zone

        data_frame = data_frame.tz_convert(alt)

        return data_frame

    def localize_index_as_UTC(self, data_frame):
        data_frame = data_frame.tz_localize(pytz.utc)

        return data_frame

    def localize_index_as_new_york_time(self, data_frame):
        new_york = pytz.timezone('America/New_York')
        data_frame = data_frame.tz_localize(new_york)

        return data_frame

    def localize_index_as_chicago_time(self, data_frame):
        chicago = pytz.timezone('America/Chicago')
        data_frame = data_frame.tz_localize(chicago)

        return data_frame

    def localize_index_as_london_time(self, data_frame):
        london = pytz.timezone('Europe/London')
        data_frame = data_frame.tz_localize(london)

        return data_frame

    def set_as_no_timezone(self, data_frame):
        data_frame = data_frame.tz_localize(None)
        return data_frame

    def tz_UTC_to_naive(self, data_frame):
        """Converts a tz-aware DatetimeIndex into a tz-naive DatetimeIndex,
        effectively baking the timezone into the internal representation.

        Parameters
        ----------
        datetime_index : pandas.DatetimeIndex, tz-aware

        Returns
        -------
        pandas.DatetimeIndex, tz-naive
        """

        # data_frame = tsc.convert_index_aware_to_UTC_time(data_frame)

        datetime_index = data_frame.index

        # Calculate timezone offset relative to UTC
        timestamp = datetime_index[0]

        tz_offset = (timestamp.replace(tzinfo=None) -
                     timestamp.tz_convert('UTC').replace(tzinfo=None))

        tz_offset_td64 = numpy.timedelta64(tz_offset)

        # Now convert to naive DatetimeIndex
        data_frame.index = pandas.DatetimeIndex(datetime_index.values + tz_offset_td64)

        return None #data_frame #(doesn't work)

    def tz_strip(self, data_frame):
        """Converts a tz-aware DatetimeIndex into a tz-naive DatetimeIndex,
        effectively baking the timezone into the internal representation.

        Parameters
        ----------
        datetime_index : pandas.DatetimeIndex, tz-aware

        Returns
        -------
        pandas.DatetimeIndex, tz-naive
        """

        # data_frame = tsc.convert_index_aware_to_UTC_time(data_frame)

        datetime_index = data_frame.index

        # Now convert to naive DatetimeIndex
        data_frame.index = pandas.DatetimeIndex(datetime_index.values)

        return None #(TODO fix as doesn't work)





__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2021 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.timeseries import Filter, Calendar, Calculations

    import pandas as pd

    calculations = Calculations()
    calendar = Calendar()
    filter = Filter()

    # choose run_example = 0 for everything
    # run_example = 1 - combine intraday dataframe with daily data dataframe

    run_example = 0

    if run_example == 1 or run_example == 0:

        df_intraday = pd.DataFrame(index=pd.date_range(start='01 Jan 2020', end='10 Jan 2020', freq='1min'),
                                   columns=['ones'])
        df_intraday['ones'] = 1

        df_intraday = df_intraday.tz_localize('utc')

        df_daily = pd.DataFrame(index=pd.date_range(start='01 Jan 2020', end='10 Jan 2020', freq='B'),
                                   columns=['ones', 'twos'])

        df_daily['ones'] = 1
        df_daily['twos'] = 2

        df_joined = calculations.join_intraday_daily(df_intraday, df_daily, daily_time_of_day='10:00', daily_time_zone='Europe/London')

        print(df_joined)









__author__ = 'saeedamen'

from findatapy.util.cachemanager import CacheManager
from findatapy.util.commonman import CommonMan
from findatapy.util.configmanager import ConfigManager
from findatapy.util.dataconstants import DataConstants
from findatapy.util.fxconv import FXConv
from findatapy.util.loggermanager import LoggerManager
from findatapy.util.singleton import Singleton
from findatapy.util.tickerfactory import TickerFactory
from findatapy.util.twitter import Twitter
from findatapy.util.swimpool import SwimPool


__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#the License for the specific language governing permissions and limitations under the License.
#

import datetime
import functools
import math

import numpy
import pandas
import pandas.tseries.offsets

from findatapy.timeseries.filter import Filter, Calendar

from pandas import compat

class DataQuality(object):
    """Checks the data quality of a DataFrame, reporting statistics such as the percentage of NaN values by column and
    through the whole DataFrame. Can also check by column between specific dates.

    """

    def percentage_nan(self, df, start_date = None):
        """Calculates the percentage of NaN values in a DataFrame.

        Parameters
        ----------
        df : DataFrame
            The data to be checked for data integrity
        start_date : str
            Filters time series by start date

        Returns
        -------
        float
            Between 0 and 100 representing the number of NaN values in a DataFrame (0 if the dataframe is None)
        """

        if df is None:
            return 100.0

        if start_date is not None:
            df = df[df.index >= start_date]

        nan = float(df.isnull().sum().sum())

        valid = float(df.count().sum())
        total = nan + valid

        if total == 0: return 0

        return round(100.0 * (nan / total), 1)

    def percentage_nan_by_columns(self, df, start_date = None):
        """Calculates the percentage of NaN values in a DataFrame and reports results by column. Likely, can do this
        for whole dataframe in one operation.

        Parameters
        ----------
        df : DataFrame
            The data to be checked for data integrity
        start_date : str
            Filters time series by start date

        Returns
        -------
        dict
            Dictionary of column names and percentage of NaN etween 0 and 100 representing the number of NaN values in each column
        """

        if start_date is not None:
            df = df[df.index >= start_date]

        nan_dict = {}

        for c in df.columns:
            nan_dict = self.percentage_nan(df[c])

        return nan_dict

    def percentage_nan_between_start_finish_dates(self, df, df_properties, asset_field, start_date_field, finish_date_field):
        """Calculates the percentage of NaN in a DataFrame in a customisable way. For each column it will only check the
        NaNs between specific start and finish dates.

        Parameters
        ----------
        df : DataFrame
            Data to be checked for integrity
        df_properties : DataFrame
            Record of each column and the start/finish dates that will be used for NaNs
        asset_field : str
            The column in df_properties which contains the column names
        start_date_field : str
            The column in df_properties which contains the start date
        finish_date_field : str
            The column in df properties which contains the finish date

        Returns
        -------
        dict
            Contains column names and the associated percentage of NaNs
        """
        percentage_nan = {}

        df_properties = df_properties.sort_values(asset_field)

        df_dates = pandas.DataFrame(index=df_properties[asset_field].values, data=df_properties[[start_date_field, finish_date_field]].values,
                                    columns=[start_date_field, finish_date_field])

        c_new = [x.split(".")[0] for x in df.columns]

        index = df_dates.index.searchsorted(c_new)
        start_date = df_dates[start_date_field][index]
        finish_date = df_dates[finish_date_field][index]

        for i in range(0, len(df.columns)):
            df_sub = df[df.columns[i]]

            percentage_nan[df.columns[i]] = self.percentage_nan(df_sub[start_date[i]:finish_date[i]])

        return percentage_nan

    def strip_dataframe_before_large_nan_section(self, df, freq = 'daily', max_nan_gap = 20):
        """For each column in a dataframe, where there is a large gap (eg. 20 working days), all values before that
        will be filled with NaN. This can for example be useful if we are constructing futures continuous time series
        and we need a continual array of futures values to be available to back adjust.

        Parameters
        ----------
        df : DataFrame
            Data to be assessed
        freq : str
            'daily' or 'intraday'
        max_nan_gap : int
            Number of days observations for missing threshold

        Returns
        -------
        DataFrame
            Overwritten earlier values with NaN, if they are before a large NaN section
        """

        if freq == 'daily':
            # resample by business days
            df_re = df.resample('B').mean()

            # calculate the rolling valid business days
            df_re_nan_count = df_re.rolling(window = max_nan_gap).count()

            # if 20 business days in a row are invalid, then flag this
            strip = df_re_nan_count[df_re_nan_count == 0]

            # for each column get the make everything before the last nan section, as equal to nan
            # eg. if 2003 is all nan, but pre 2003 is populated by real values, we'll overwrite the pre 2003 values with
            # nan, can cause issues with backtesting if we have large nan sections
            for c in df.columns:
                strip_index = strip[c].last_valid_index()

                # only overwrite, if we have poor quality sections (if the data quality is good we won't have this issue)
                if strip_index is not None:
                    df.ix[:strip_index,c] = numpy.nan

        elif freq == 'intraday':
            pass

        return df

    def count_repeated_dates(self, df):
        """Counts number of duplicated dates in a DataFrame and returns these

        Parameters
        ----------
        df : DataFrame
            Data to be checked

        Returns
        -------
        int, DateTimeIndex
            Number of duplicated entries, the duplicated entries themselves
        """

        duplicated = df.index.duplicated()

        return len(duplicated), df.index[duplicated]




__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    from findatapy.util.dataconstants import DataConstants

    eikon_api_key = DataConstants().eikon_api_key

    df = None

    if eikon_api_key is None:
        eikon_api_key = 'TYPE_YOUR_API'

    import datetime
    from datetime import timedelta

    # You need to have Eikon installed and to have a valid licence for this to work
    # For intraday pricing, you can usually access a history back a few months from the current date
    # (if you need older history there are other Refinitiv products like Tick History)

    # Note: can sometimes get following error
    # eikon.eikonError.EikonError: Error code -1 | Port number was not identified. Check if Eikon Desktop or Eikon API Proxy is running.

    # Steps to check if you get the above error when trying to download via the Eikon Python package, which findatapy uses
    #
    # 1. Check Eikon Desktop is running, also sometimes downgrading your Eikon Python package might help
    # 2. Usually Eikon Data API is running on port 9000, to check this is running try accessing page
    # http://localhost:9000/ping?all
    #
    # The output should be something like this, if the Data API is running
    # {"port":9000,"mode":"eikon4","pid":23008,"hasSecure":true,"startedTime":"Wed Sep 09 2020 16:06:24 GMT+0100 (GMT Daylight Time)",
    # "subApps":[{"path":"/heap"},{"path":"/ping"},{"path":"/sxs","data":{"hasSecure":true,"sxsApps":{}}},{"path":"/api"}]}
    #
    # 3. In folder C:\Users\YOURUSERNAME\AppData\Roaming\Refinitiv\Data API Proxy there should be a file .portInUse which has
    # only the port used by Eikon eg. the below (if this is different to the port which the Data API Proxy is running, you'll
    # likely get the above error)
    #
    # 9000
    #
    # 4. On older versions of Eikon this folder might be c:\Users\YOURUSERNAME\AppData\Thomson Reuters\Eikon API Proxy\
    # 5. It might be necessary to check Eikon logs, to see which port the Data API is running on too, if it's not running on 9000
    # 6. Also check various Refinitiv forum pages such as
    # https://community.developers.refinitiv.com/questions/37673/app-cannot-connect-to-server-port-9000-target-mach.html

    # Get recent tick data for FX tick data
    md_request = MarketDataRequest(
        start_date=datetime.datetime.utcnow() - timedelta(hours=1),     # Start date (download data over past decade)
        freq='tick',
        data_source='eikon',                                            # Use Eikon as data source
        tickers=['EURUSD'],                                             # ticker
        fields=['bid', 'ask'],                                          # which fields to download
        vendor_tickers=['EUR='],                                        # ticker (Eikon/RIC)
        vendor_fields=['BID', 'ASK'],                                   # Which Eikon fields to download
        eikon_api_key=eikon_api_key)

    df = market.fetch_market(md_request)

    print(df)

    # Get 1 minute intraday data for FX
    md_request = MarketDataRequest(
        start_date=datetime.datetime.utcnow() - timedelta(days=5),    # Start date (download data over past decade)
        freq='intraday',
        data_source='eikon',                                           # use Eikon as data source
        tickers=['EURUSD', 'GBPUSD'],                                  # ticker
        fields=['close', 'open', 'high', 'low'],                       # which fields to download
        vendor_tickers=['EUR=', 'GBP='],                               # ticker (Eikon/RIC)
        vendor_fields=['CLOSE', 'OPEN', 'HIGH', 'LOW'],                # which Eikon fields to download
        eikon_api_key=eikon_api_key)

    df = market.fetch_market(md_request)

    print(df)

    # We can also download G10 FX from Eikon using shortcuts, without specifying RIC
    # we just need to specify category at 'fx'
    # You can add your own customized tickers by editing the various conf CSV files
    md_request = MarketDataRequest(
        start_date=datetime.datetime.utcnow() - timedelta(days=5),    # Start date (download data over past decade)
        freq='intraday',
        category='fx',
        cut='NYC',
        data_source='eikon',                                           # use Eikon as data source
        tickers=['EURUSD', 'GBPUSD'],                                  # ticker
        fields=['close', 'open', 'high', 'low'],                       # which fields to download
        eikon_api_key=eikon_api_key)

    print(df)

    df = market.fetch_market(md_request)

    # Also let's do this for daily data
    md_request.freq = 'daily'

    df = market.fetch_market(md_request)

    print(df)


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download FX forwards curve data from Bloomberg for FX (and FX spot and base depos) for EURUSD
    # run_example = 2 - download FX forwards curve data from Bloomberg for FX (and FX spot and base depos) for USDBRL

    run_example = 0

    cache_algo = 'cache_algo_return'

    ###### Download FX forwards data for EURUSD
    if run_example == 1 or run_example == 0:

        ####### Bloomberg examples (you need to have a Bloomberg Terminal installed for this to work!)

        # We can use shortcuts given that forwards for most major crosses have been defined
        # Get EURUSD 1M forward points
        md_request = MarketDataRequest(start_date='01 Oct 2020', finish_date='18 Dec 2020',
                                       data_source='bloomberg', cut='NYC', category='fx-forwards',
                                       tickers=['EURUSD1M'], cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # Get whole of EURUSD FX forwards curve + spot + base depos
        md_request = MarketDataRequest(start_date='01 Oct 2020', finish_date='18 Dec 2020',
                                       data_source='bloomberg', cut='NYC', category='fx-forwards-market',
                                       tickers=['EURUSD'], cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

    ###### Download FX forwards data for USDBRL
    if run_example == 2 or run_example == 0:

        ####### Bloomberg examples (you need to have a Bloomberg Terminal installed for this to work!)

        # We can use shortcuts given that forwards for most major crosses have been defined
        # Get USDBRL 3M forward points
        md_request = MarketDataRequest(start_date='01 Oct 2020', finish_date='18 Dec 2020',
                                       data_source='bloomberg', cut='NYC', category='fx-forwards',
                                       tickers=['USDBRL3M'], cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # Get whole of EURUSD FX forwards curve + spot + base depos
        md_request = MarketDataRequest(start_date='01 Oct 2020', finish_date='18 Dec 2020',
                                       data_source='bloomberg', cut='NYC', category='fx-forwards-market',
                                       tickers=['USDBRL'], cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))




__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    md_request = MarketDataRequest(
        start_date="01 Jun 2000",                   # start date (download data over past decade)
        data_source='alfred',                       # use ALFRED/FRED as data source
        tickers=['US CPI YoY', 'EZ CPI YoY'],       # ticker
        fields=['close'],                           # which fields to download
        vendor_tickers=['CPIAUCSL', 'CP0000EZ17M086NEST'],  # ticker (FRED)
        vendor_fields=['close'])                    # which ALFRED fields to download

    df = market.fetch_market(md_request)

    print(df.tail(n=10))


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - read in CSV file (daily)
    # run_example = 2 - read in HD5 file in market data (intraday)
    # run_example = 3 - save to disk for Quandl as Parquet (defined tickers) and read back with MarketDataRequest
    # run_example = 4 - save to disk for Dukascopy as Parquet (defined tickers for EURUSD) and read back with MarketDataRequest

    # NOTE: you need to make sure you have the correct data licences before storing data on disk (and whether other
    # users can access it)

    run_example = 4

    if run_example == 1 or run_example == 0:

        md_request = MarketDataRequest(
            start_date='01 Jan 2002', finish_date='31 Jan 2016',
            tickers='S&P500',
            fields='close',
            data_source="../tests/S&P500.csv",
            freq='daily',
        )

        market = Market(market_data_generator=MarketDataGenerator())

        df = market.fetch_market(md_request=md_request)

        print(df)

    if run_example == 2 or run_example == 0:
        # load tick data from DukasCopy and then resample to 15s buckets
        md_request = MarketDataRequest(
            start_date='01 Jun 2016', finish_date='31 Jul 2016',
            tickers='EURUSD',
            fields='bid',
            data_source="../tests/EURUSD_tick.h5",
            freq='intraday',
            resample='15s',
            resample_how='last_dropna'
        )

        market = Market(market_data_generator=MarketDataGenerator())

        df = market.fetch_market(md_request=md_request)

        print(df)

        # the second time we call it, if we have Redis installed, we will fetch from memory, so it will be quicker
        # also don't need to run the resample operation again

        # need to specify cache_algo_return
        md_request.cache_algo = 'cache_algo_return'

        df = market.fetch_market(md_request)

        print(df)

    if run_example == 3:
        # In this case we are saving predefined daily tickers to disk, and then reading back
        from findatapy.util.dataconstants import DataConstants
        from findatapy.market.ioengine import IOEngine
        import os

        quandl_api_key = DataConstants().quandl_api_key # change with your own Quandl API key!

        md_request = MarketDataRequest(
            category='fx',
            data_source='quandl',
            freq='daily',
            quandl_api_key=quandl_api_key
        )

        market = Market(market_data_generator=MarketDataGenerator())

        df = market.fetch_market(md_request=md_request)

        print(df)

        folder = '../tests/'

        # Save to disk in a file name format friendly for reading later via MarketDataRequest (ie. ../tests/backtest.fx.daily.quandl.NYC.parquet)
        IOEngine().write_time_series_cache_to_disk(folder, df, engine='parquet', md_request=md_request)

        md_request.data_engine = '../tests/*.parquet'

        df = market.fetch_market(md_request)

        print(df)

    if run_example == 4:
        # In this case we are saving predefined tick data tickers to disk, and then reading back using the MarketDataRequest interface
        from findatapy.util.dataconstants import DataConstants
        from findatapy.market.ioengine import IOEngine

        md_request = MarketDataRequest(
            start_date='01 Jan 2021',
            finish_date='05 Jan 2021',
            category='fx',
            data_source='dukascopy',
            freq='tick',
            tickers=['EURUSD'],
            fields=['bid', 'ask', 'bidv', 'askv'],
        )

        market = Market(market_data_generator=MarketDataGenerator())

        df = market.fetch_market(md_request=md_request)

        print(df)

        folder = '../tests/'

        # Save to disk in a format friendly for reading later (ie. ../tests/backtest.fx.tick.dukascopy.NYC.EURUSD.parquet)
        # Here it will automatically generate the filename from the folder we gave
        # and the MarketDataRequest we made (altenatively, we could have just given the filename directly)
        IOEngine().write_time_series_cache_to_disk(folder, df, engine='parquet', md_request=md_request)

        md_request.data_engine = '../tests/*.parquet'
        df = market.fetch_market(md_request)

        print(df)

        # Let's try the same thing with CSV too! But this will be slower given it's CSV format...!
        IOEngine().write_time_series_cache_to_disk(folder, df, engine='csv', md_request=md_request)

        md_request.data_engine = '../tests/*.csv'
        df = market.fetch_market(md_request)

        print(df)


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2021 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.timeseries import Filter, Calendar

    import pandas as pd
    import numpy as np

    calendar = Calendar()
    filter = Filter()

    # choose run_example = 0 for everything
    # run_example = 1 - get holidays for FX, EUR and EURUSD, as well as listing weekends
    # run_example = 2 - get FX delivery dates and FX option expiries for various tenors
    # run_example = 3 - get number of days between pandas DatetimeIndex
    # run_example = 4 - filter time series by EURUSD holidays
    # run_example = 4 - option expiries for USDJPY

    run_example = 0

    if run_example == 1 or run_example == 0:

        # Get the holidays (which aren't weekends)
        print(calendar.get_holidays(start_date='01 Jan 1999 00:50', end_date='31 Dec 1999', cal='FX'))
        print(calendar.get_holidays(start_date='01 Jan 2000 00:10', end_date='31 Dec 2000', cal='EUR'))
        print(calendar.get_holidays(start_date='01 Jan 2000 00:10', end_date='31 Dec 2000', cal='EURUSD'))

        # Get the holidays (which are weekends)
        print(calendar.get_holidays(start_date='01 Jan 1999 00:50', end_date='31 Dec 1999', cal='WKY'))

    if run_example == 2 or run_example == 0:

        # Get delivery dates for these horizon dates - typically would use to get forward maturities
        print(calendar.get_delivery_date_from_horizon_date(
            pd.to_datetime([pd.Timestamp('02 Nov 2020')]), 'ON', cal='EURUSD'))

        print(calendar.get_delivery_date_from_horizon_date(
            pd.to_datetime([pd.Timestamp('02 Nov 2020')]), '1W', cal='EURUSD'))

        print(calendar.get_delivery_date_from_horizon_date(
            pd.to_datetime([pd.Timestamp('27 Nov 2020')]), '1W', cal='EURUSD'))

        # Get 1M expires for these horizon dates - typically would use to get option expiries
        print(calendar.get_expiry_date_from_horizon_date(
            pd.to_datetime([pd.Timestamp('26 Oct 2020')]), '1M', cal='EURUSD'))

    if run_example == 3 or run_example == 0:
        # Create a list of business days and one which is + 1 day

        bus_days = pd.bdate_range('1 Jan 2020', '30 Jan 2020')
        bus_days_plus = bus_days + pd.Timedelta(days=7)

        print(calendar.get_delta_between_dates(bus_days, bus_days_plus))

    if run_example == 4 or run_example == 0:
        # Filter a time series by EURUSD holidays

        df = pd.DataFrame(index=pd.bdate_range('1 Jan 2020', '31 Dec 2020'))
        df['Prices'] = 1

        df_filtered = filter.filter_time_series_by_holidays(df, 'EURUSD')

        print(len(df.index))
        print(len(df_filtered.index))

    if run_example == 5 or run_example == 0:
        # Get expiry for USDJPY

        # Get 1M expires for these horizon dates - typically would use to get option expiries
        print(calendar.get_expiry_date_from_horizon_date(
            pd.to_datetime([pd.Timestamp('26 Nov 2008')]), '1M', cal='USDJPY'))






__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # Get the first release for GDP and also print the release date of that
    md_request = MarketDataRequest(
        start_date="01 Jun 2000",                                                      # start date (download data over past decade)
        data_source='alfred',                                                          # use ALFRED/FRED as data source
        tickers=['US GDP'],                                                            # ticker
        fields=['actual-release', 'release-date-time-full'],                           # which fields to download
        vendor_tickers=['GDP'],                                                        # ticker (FRED)
        vendor_fields=['actual-release', 'release-date-time-full'])                    # which FRED fields to download

    df = market.fetch_market(md_request)

    print(df)

    # Compare the close and actual release of US GDP (and the final)
    md_request = MarketDataRequest(
        start_date="01 Jun 2000",                                                      # start date (download data over past decade)
        data_source='alfred',                                                          # use ALFRED/FRED as data source
        tickers=['US GDP'],                                                            # ticker
        fields=['actual-release', 'close'],                                            # which fields to download
        vendor_tickers=['GDP'],                                                        # ticker (FRED)
        vendor_fields=['actual-release', 'close'])                                     # which FRED fields to download

    df = market.fetch_market(md_request)

    from chartpy import Chart, Style

    style = Style(title="US GDP first vs last")

    Chart().plot(df, style=style)

    # Get the change NFP SA (need to calculate that from the acutal-release and first-revision)
    md_request = MarketDataRequest(
        start_date="01 Jun 2000",                                                      # start date (download data over past decade)
        data_source='alfred',                                                          # use ALFRED/FRED as data source
        tickers=['US NFP'],                                                            # ticker
        fields=['actual-release', 'first-revision', 'release-date-time-full'],         # which fields to download
        vendor_tickers=['PAYEMS'],                                                     # ticker (FRED)
        vendor_fields=['actual-release', 'first-revision', 'release-date-time-full'])  # which FRED fields to download

    df = market.fetch_market(md_request)

    # calculate the headline change in NFP
    df['US NFP change'] = df['US NFP.actual-release'] - df['US NFP.first-revision'].shift(1)

    print(df)

    from chartpy import Chart, Style
    import pandas

    style = Style(title="US NFP change (actual)")

    df1 = pandas.DataFrame(df['US NFP change'])

    Chart().plot(df1, style=style)

    # Get release times on their own
    # Get the change NFP SA
    # need to calculate that from the acutal-release and first-revision)
    md_request = MarketDataRequest(
        start_date="01 Aug 2013",
        finish_date="30 Nov 2019",
        data_source='alfred',
        tickers=['US NFP'],
        fields=['release-date-time-full'],
        vendor_tickers=['PAYEMS'],
        vendor_fields=['release-date-time-full'])

    market = Market(market_data_generator=MarketDataGenerator())

    df_nfp = market.fetch_market(md_request)

    print(df_nfp)



from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
ext_modules = [Extension('bbgloop',
                         ['bbgloop.pyx'],
                        )]
setup(
name = 'bbgloop',
cmdclass = {'build_ext': build_ext},
ext_modules = ext_modules
)

# python setup.py build_ext -i clean


__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

from findatapy.util.loggermanager import LoggerManager

class FXConv(object):
    """Various methods to manipulate FX crosses, applying correct conventions.

    """

    # TODO
    g10 = ['EUR', 'GBP', 'AUD', 'NZD', 'USD', 'CAD', 'CHF', 'NOK', 'SEK', 'JPY']

    order = ['XBT', 'XET', 'XLC', 'XRP', 'XAU', 'XPT', 'XAG', 'EUR', 'GBP', 'AUD', 'NZD', 'USD', 'CAD', 'CHF',
             'TRY', 'SGD', 'CHN', 'CNY',
             'NOK', 'SEK',
             'BRL', 'ZAR', 'MXN',
             'TWD', 'INR',
             'JPY',
             'CLP', 'COP', 'KRW', 'IDR']

    ndf = ['CNY', 'IDR', 'INR', 'KRW', 'MYR', 'PHP', 'TWD', 'VND',
           'BRL', 'ARS', 'BRL', 'CLP', 'GTQ', 'PEN', 'UVU', 'VEF'
           'EGP', 'KZT', 'NGN']

    def __init__(self):
        self.logger = LoggerManager().getLogger(__name__)
        return

    def g10_crosses(self):

        g10_crosses = []

        for i in range(0, len(self.g10)):
            for j in range(0, len(self.g10)):
                if i != j:
                    g10_crosses.append(self.correct_notation(self.g10[i] + self.g10[j]))

        set_val = set(g10_crosses)
        g10_crosses = sorted(list(set_val))

        return g10_crosses

    def em_or_g10(self, currency, freq = "daily"):
        if freq == 'intraday':
            return 'fx'

        try:
            index = self.g10.index(currency)
        except ValueError:
            index = -1

        if (index < 0):
            return 'fx-em'

        return 'fx-g10'

    def is_USD_base(self, cross):
        base = cross[0:3]
        terms = cross[3:6]

        if base == 'USD':
            return True

        return False

    def is_EM_cross(self, cross):
        base = cross[0:3]
        terms = cross[3:6]

        if self.em_or_g10(base, 'daily') == 'fx-em' or self.em_or_g10(terms, 'daily') == 'fx-em':
            return True

        return False

    def is_NDF_cross(self, cross):
        base = cross[0:3]
        terms = cross[3:6]

        if base in self.ndf or terms in self.ndf:
            return True

        return False

    def decompose_pair_into_USD_crosses(self, cross):
        base = cross[0:3]
        terms = cross[3:6]

        if base == 'USD' or terms == 'USD':
            return cross, 'USDUSD'

        return base + 'USD', 'USD' + terms

    def correct_notation(self, cross):
        base = cross[0:3]
        terms = cross[3:6]

        try:
            base_index = self.order.index(base)
        except ValueError:
            base_index = -1

        try:
            terms_index = self.order.index(terms)
        except ValueError:
            terms_index = -1

        if (base_index < 0 and terms_index > 0):
            return terms + base
        if (base_index > 0 and terms_index < 0):
            return base + terms
        elif (base_index > terms_index):
            return terms + base
        elif (terms_index > base_index):
            return base + terms

        return cross

if __name__ == '__main__':
    logger = LoggerManager.getLogger(__name__)

    fxconv = FXConv()

    if True:
        logger.info(fxconv.g10_crosses())



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

from twython import Twython

from findatapy.util import DataConstants, LoggerManager

class Twitter(object):
    """Has functions to tweet from Python (using Twython library)

    """

    def __init__(self, *args, **kwargs):
        self.logger = LoggerManager().getLogger(__name__)

    def set_key(self, APP_KEY, APP_SECRET, OAUTH_TOKEN, OAUTH_TOKEN_SECRET):
        self.twitter = Twython(APP_KEY, APP_SECRET, OAUTH_TOKEN, OAUTH_TOKEN_SECRET)

    def auto_set_key(self):
        self.twitter = Twython(DataConstants().APP_KEY, DataConstants().APP_SECRET,
                               DataConstants().OAUTH_TOKEN, DataConstants().OAUTH_TOKEN_SECRET)

    def update_status(self, msg, link = None, picture = None):
        # 22 chars URL
        # 23 chars picture

        chars_lim = 140

        if link is not None: chars_lim = chars_lim - (22 * link)
        if picture is not None: chars_lim = chars_lim - 23

        if (len(msg) > chars_lim):
            self.logger.info("Message too long for Twitter!")

        if picture is None:
            self.twitter.update_status(status=msg)
        else:
            photo = open(picture, 'rb')
            self.twitter.update_status_with_media(status=msg, media=photo)







__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import numpy as np
import codecs
import glob
import datetime
from dateutil.parser import parse
import shutil
import copy

import math

try:
    import bcolz
except:
    pass

# Needs this for AWS S3 bucket support
try:
    from s3fs import S3FileSystem
except:
    pass

# pyarrow necessary for caching
try:
    import pyarrow as pa
except:
    pass

# for reading and writing to S3
try:
    import pyarrow.fs
    import pyarrow.parquet as pq

    from s3fs import S3FileSystem
except:
    pass

try:
    import redis
except:
    pass

import pandas as pd

from openpyxl import load_workbook
import os.path

from findatapy.util.dataconstants import DataConstants
from findatapy.util.loggermanager import LoggerManager

# NOTE: BCOLZ support is alpha!
_replace_chars = ['_a_',
                  '_d_',
                  '_h_',
                  '_o_',
                  '_c_',
                  '_s_',
                  '_p_',
                  '_e_',
                  '_'
                  ]

_invalid_chars = ['&',
                  '.',
                  '-',
                  '(',
                  ')',
                  '/',
                  '%',
                  '=',
                  ' ']

constants = DataConstants()

class IOEngine(object):
    """Write and reads time series data to disk in various formats, CSV, HDF5 (fixed and table formats) and MongoDB/Arctic.

    Can be used to save down output of finmarketpy backtests and also to cache market data locally.

    Also supports BColz (but not currently stable). Planning to add other interfaces such as SQL etc.

    """

    def __init__(self):
        pass

    ### functions to handle Excel on disk
    def write_time_series_to_excel(self, fname, sheet, data_frame, create_new=False):
        """Writes Pandas data frame to disk in Excel format

        Parameters
        ----------
        fname : str
            Excel filename to be written to
        sheet : str
            sheet in excel
        data_frame : DataFrame
            data frame to be written
        create_new : boolean
            to create a new Excel file
        """

        if (create_new):
            writer = pd.ExcelWriter(fname, engine='xlsxwriter')
        else:
            if self.path_exists(fname):
                book = load_workbook(fname)
                writer = pd.ExcelWriter(fname, engine='xlsxwriter')
                writer.book = book
                writer.sheets = dict((ws.title, ws) for ws in book.worksheets)
            else:
                writer = pd.ExcelWriter(fname, engine='xlsxwriter')

        data_frame.to_excel(writer, sheet_name=sheet, engine='xlsxwriter')

        writer.save()
        writer.close()

    def write_time_series_to_excel_writer(self, writer, sheet, data_frame):
        """Writes Pandas data frame to disk in Excel format for a writer

        Parameters
        ----------
        writer : ExcelWriter
            File handle to use for writing Excel file to disk
        sheet : str
            sheet in excel
        data_frame : DataFrame
            data frame to be written
        """
        data_frame.to_excel(writer, sheet, engine='xlsxwriter')

    def read_excel_data_frame(self, f_name, excel_sheet, freq, cutoff=None, dateparse=None,
                              postfix='.close', intraday_tz='UTC'):
        """Reads Excel from disk into DataFrame

        Parameters
        ----------
        f_name : str
            Excel file path to read
        freq : str
            Frequency of data to read (intraday/daily etc)
        cutoff : DateTime (optional)
            end date to read up to
        dateparse : str (optional)
            date parser to use
        postfix : str (optional)
            postfix to add to each columns
        intraday_tz : str
            timezone of file if uses intraday data

        Returns
        -------
        DataFrame
        """

        return self.read_csv_data_frame(f_name, freq, cutoff=cutoff, dateparse=dateparse,
                                        postfix=postfix, intraday_tz=intraday_tz, excel_sheet=excel_sheet)

    def remove_time_series_cache_on_disk(self, fname, engine='hdf5_fixed', db_server='127.0.0.1', db_port='6379',
                                         timeout=10, username=None,
                                         password=None):

        logger = LoggerManager().getLogger(__name__)

        if 'hdf5' in engine:
            engine = 'hdf5'

        if (engine == 'bcolz'):
            # convert invalid characters to substitutes (which Bcolz can't deal with)
            pass
        elif (engine == 'redis'):

            fname = os.path.basename(fname).replace('.', '_')

            try:
                r = redis.StrictRedis(host=db_server, port=db_port, db=0, socket_timeout=timeout,
                                      socket_connect_timeout=timeout)

                if (fname == 'flush_all_keys'):
                    r.flushall()
                else:
                    # allow deletion of keys by pattern matching

                    x = r.keys('*' + fname)

                    if len(x) > 0:
                        r.delete(x)

                    # r.delete(fname)

            except Exception as e:
                logger.warning("Cannot delete non-existent key " + fname + " in Redis: " + str(e))

        elif (engine == 'arctic'):
            from arctic import Arctic
            import pymongo

            socketTimeoutMS = 30 * 1000
            fname = os.path.basename(fname).replace('.', '_')

            logger.info('Load MongoDB library: ' + fname)

            if username is not None and password is not None:
                c = pymongo.MongoClient(
                    host="mongodb://" + username + ":" + password + "@" + str(db_server) + ":" + str(db_port),
                    connect=False)  # , username=username, password=password)
            else:
                c = pymongo.MongoClient(host="mongodb://" + str(db_server) + ":" + str(db_port), connect=False)

            store = Arctic(c, socketTimeoutMS=socketTimeoutMS, serverSelectionTimeoutMS=socketTimeoutMS,
                           connectTimeoutMS=socketTimeoutMS)

            store.delete_library(fname)

            c.close()

            logger.info("Deleted MongoDB library: " + fname)

        elif (engine == 'hdf5'):
            h5_filename = self.get_h5_filename(fname)

            # delete the old copy
            try:
                os.remove(h5_filename)
            except:
                pass

    ### functions to handle HDF5 on disk, arctic etc.
    def write_time_series_cache_to_disk(self, fname, data_frame,
                                        engine='hdf5_fixed', append_data=False, db_server=constants.db_server,
                                        db_port=constants.db_port, username=constants.db_username, password=constants.db_password,
                                        filter_out_matching=None, timeout=10,
                                        use_cache_compression=constants.use_cache_compression,
                                        parquet_compression=constants.parquet_compression, use_pyarrow_directly=False,
                                        md_request=None, ticker=None, cloud_credentials=None):
        """Writes Pandas data frame to disk as HDF5 format or bcolz format or in Arctic

        Parmeters
        ---------
        fname : str
            path of file
        data_frame : DataFrame
            data frame to be written to disk
        engine : str
            'hdf5_fixed' - use HDF5 fixed format, very quick, but cannot append to this
            'hdf5_table' - use HDF5 table format, slower but can append to
            'parquet' - use Parquet
            'arctic' - use Arctic/MongoDB database
            'redis' - use Redis
        append_data : bool
            False - write a fresh copy of data on disk each time
            True - append data to disk
        db_server : str
            Database server for arctic (default: '127.0.0.1')
        timeout : int
            Number of seconds to do timeout
        """

        logger = LoggerManager().getLogger(__name__)

        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if md_request is not None:
            fname = self.path_join(fname, md_request.create_category_key(ticker=ticker))

        # default HDF5 format
        hdf5_format = 'fixed'

        if 'hdf5' in engine:
            hdf5_format = engine.split('_')[1]
            engine = 'hdf5'

        if (engine == 'bcolz'):
            # convert invalid characters to substitutes (which Bcolz can't deal with)
            data_frame.columns = self.find_replace_chars(data_frame.columns, _invalid_chars, _replace_chars)
            data_frame.columns = ['A_' + x for x in data_frame.columns]

            data_frame['DTS_'] = pd.to_datetime(data_frame.index, unit='ns')

            bcolzpath = self.get_bcolz_filename(fname)
            shutil.rmtree(bcolzpath, ignore_errors=True)
            zlens = bcolz.ctable.fromdataframe(data_frame, rootdir=bcolzpath)
        elif (engine == 'redis'):

            fname = os.path.basename(fname).replace('.', '_')

            # Will fail if Redis is not installed
            try:
                r = redis.StrictRedis(host=db_server, port=db_port, db=0, socket_timeout=timeout,
                                      socket_connect_timeout=timeout)

                ping = r.ping()

                # If Redis is alive, try pushing to it
                if ping:
                    if data_frame is not None:
                        if isinstance(data_frame, pd.DataFrame):
                            mem = data_frame.memory_usage(deep='deep').sum()
                            mem_float = round(float(mem) / (1024.0 * 1024.0), 3)

                            if mem_float < 500:
                                # msgpack/blosc is deprecated
                                # r.set(fname, data_frame.to_msgpack(compress='blosc'))

                                # Now uses pyarrow
                                try:
                                    context = pa
                                    ser = context.serialize(data_frame).to_buffer()
                                except:
                                    # For earlier pyarrow versions
                                    context = pa.default_serialization_context()
                                    ser = context.serialize(data_frame).to_buffer()

                                if use_cache_compression:
                                    comp = pa.compress(ser, codec='gzip', asbytes=True)
                                    siz = len(ser)  # siz = 3912

                                    r.set('comp_' + str(siz) + '_' + fname, comp)
                                else:
                                    r.set(fname, ser.to_pybytes())

                                logger.info("Pushed " + fname + " to Redis")
                            else:
                                logger.warn("Did not push " + fname + " to Redis, given size")
                    else:
                        logger.info("Object " + fname + " is empty, not pushed to Redis.")
                else:
                    logger.warning("Didn't push " + fname + " to Redis given not running")

            except Exception as e:
                logger.warning("Couldn't push " + fname + " to Redis: " + str(e))

        elif (engine == 'arctic'):
            from arctic import Arctic
            import pymongo

            socketTimeoutMS = 30 * 1000
            fname = os.path.basename(fname).replace('.', '_')

            logger.info('Load Arctic/MongoDB library: ' + fname)

            if username is not None and password is not None:
                c = pymongo.MongoClient(
                    host="mongodb://" + username + ":" + password + "@" + str(db_server) + ":" + str(db_port),
                    connect=False)  # , username=username, password=password)
            else:
                c = pymongo.MongoClient(host="mongodb://" + str(db_server) + ":" + str(db_port), connect=False)

            store = Arctic(c, socketTimeoutMS=socketTimeoutMS, serverSelectionTimeoutMS=socketTimeoutMS,
                           connectTimeoutMS=socketTimeoutMS)

            database = None

            try:
                database = store[fname]
            except:
                pass

            if database is None:
                store.initialize_library(fname, audit=False)
                logger.info("Created MongoDB library: " + fname)
            else:
                logger.info("Got MongoDB library: " + fname)

            # Access the library
            library = store[fname]

            if ('intraday' in fname):
                data_frame = data_frame.astype('float32')

            if filter_out_matching is not None:
                cols = data_frame.columns

                new_cols = []

                for col in cols:
                    if filter_out_matching not in col:
                        new_cols.append(col)

                data_frame = data_frame[new_cols]

            # Problems with Arctic when writing timezone to disk sometimes, so strip
            data_frame = data_frame.copy().tz_localize(None)

            try:
                # Can duplicate values if we have existing dates
                if append_data:
                    library.append(fname, data_frame)
                else:
                    library.write(fname, data_frame)

                c.close()
                logger.info("Written MongoDB library: " + fname)
            except Exception as e:
                logger.warning("Couldn't write MongoDB library: " + fname + " " + str(e))

        elif (engine == 'hdf5'):
            h5_filename = self.get_h5_filename(fname)

            # append data only works for HDF5 stored as tables (but this is much slower than fixed format)
            # removes duplicated entries at the end
            if append_data:
                store = pd.HDFStore(h5_filename, format=hdf5_format, complib="zlib", complevel=9)

                if ('intraday' in fname):
                    data_frame = data_frame.astype('float32')

                # get last row which matches and remove everything after that (because append
                # function doesn't check for duplicated rows
                nrows = len(store['data'].index)
                last_point = data_frame.index[-1]

                i = nrows - 1

                while (i > 0):
                    read_index = store.select('data', start=i, stop=nrows).index[0]

                    if (read_index <= last_point): break

                    i = i - 1

                # remove rows at the end, which are duplicates of the incoming time series
                store.remove(key='data', start=i, stop=nrows)
                store.put(key='data', value=data_frame, format=hdf5_format, append=True)
                store.close()
            else:
                h5_filename_temp = self.get_h5_filename(fname + ".temp")

                # delete the old copy
                try:
                    os.remove(h5_filename_temp)
                except:
                    pass

                store = pd.HDFStore(h5_filename_temp, complib="zlib", complevel=9)

                if ('intraday' in fname):
                    data_frame = data_frame.astype('float32')

                store.put(key='data', value=data_frame, format=hdf5_format)
                store.close()

                # delete the old copy
                try:
                    os.remove(h5_filename)
                except:
                    pass

                # once written to disk rename
                os.rename(h5_filename_temp, h5_filename)

            logger.info("Written HDF5: " + fname)

        elif (engine == 'parquet'):
            if '.parquet' not in fname:
                if fname[-5:] != '.gzip':
                    fname = fname + '.parquet'

            self.to_parquet(data_frame, fname, cloud_credentials=cloud_credentials, parquet_compression=parquet_compression,
                            use_pyarrow_directly=use_pyarrow_directly)
            # data_frame.to_parquet(fname, compression=parquet_compression)

            logger.info("Written Parquet: " + fname)
        elif engine == 'csv':
            if '.csv' not in fname:
                fname = fname + '.csv'

            data_frame.to_csv(fname)

            logger.info("Written CSV: " + fname)


    def get_h5_filename(self, fname):
        """Strips h5 off filename returning first portion of filename

        Parameters
        ----------
        fname : str
            h5 filename to strip

        Returns
        -------
        str
        """
        if fname[-3:] == '.h5':
            return fname

        return fname + ".h5"

    def get_bcolz_filename(self, fname):
        """Strips bcolz off filename returning first portion of filename

        Parameters
        ----------
        fname : str
            bcolz filename to strip

        Returns
        -------
        str
        """
        if fname[-6:] == '.bcolz':
            return fname

        return fname + ".bcolz"

    def write_r_compatible_hdf_dataframe(self, data_frame, fname, fields=None):
        """Write a DataFrame to disk in as an R compatible HDF5 file.

        Parameters
        ----------
        data_frame : DataFrame
            data frame to be written
        fname : str
            file path to be written
        fields : list(str)
            columns to be written
        """

        logger = LoggerManager().getLogger(__name__)

        fname_r = self.get_h5_filename(fname)

        logger.info("About to dump R binary HDF5 - " + fname_r)
        data_frame32 = data_frame.astype('float32')

        if fields is None:
            fields = data_frame32.columns.values

        # decompose date/time into individual fields (easier to pick up in R)
        data_frame32['Year'] = data_frame.index.year
        data_frame32['Month'] = data_frame.index.month
        data_frame32['Day'] = data_frame.index.day
        data_frame32['Hour'] = data_frame.index.hour
        data_frame32['Minute'] = data_frame.index.minute
        data_frame32['Second'] = data_frame.index.second
        data_frame32['Millisecond'] = data_frame.index.microsecond / 1000

        data_frame32 = data_frame32[
            ['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second', 'Millisecond'] + fields]

        cols = data_frame32.columns

        store_export = pd.HDFStore(fname_r)
        store_export.put('df_for_r', data_frame32, data_columns=cols)
        store_export.close()

    def read_time_series_cache_from_disk(self, fname, engine='hdf5', start_date=None, finish_date=None,
                                         db_server=constants.db_server,
                                         db_port=constants.db_port, username=constants.db_username, password=constants.db_password):
        """Reads time series cache from disk in either HDF5 or bcolz

        Parameters
        ----------
        fname : str (or list)
            file to be read from
        engine : str (optional)
            'hd5' - reads HDF5 files (default)
            'arctic' - reads from Arctic/MongoDB database
            'bcolz' - reads from bcolz file (not fully implemented)
            'parquet' - reads from Parquet
        start_date : str/datetime (optional)
            Start date
        finish_date : str/datetime (optional)
            Finish data
        db_server : str
            IP address of MongdDB (default '127.0.0.1')

        Returns
        -------
        DataFrame
        """

        logger = LoggerManager.getLogger(__name__)

        data_frame_list = []

        if not(isinstance(fname, list)):
            if '*' in fname:
                fname = glob.glob(fname)
            else:
                fname = [fname]

        for fname_single in fname:
            logger.debug("Reading " + fname_single + "..")

            if engine == 'parquet' and '.gzip' not in fname_single and '.parquet' not in fname_single:
                fname_single = fname_single + '.parquet'

            if (engine == 'bcolz'):
                try:
                    name = self.get_bcolz_filename(fname_single)
                    zlens = bcolz.open(rootdir=name)
                    data_frame = zlens.todataframe()

                    data_frame.index = pd.DatetimeIndex(data_frame['DTS_'])
                    data_frame.index.name = 'Date'
                    del data_frame['DTS_']

                    # convert invalid characters (which Bcolz can't deal with) to more readable characters for pandas
                    data_frame.columns = self.find_replace_chars(data_frame.columns, _replace_chars, _invalid_chars)
                    data_frame.columns = [x[2:] for x in data_frame.columns]
                except:
                    data_frame = None

            elif (engine == 'redis'):
                fname_single = os.path.basename(fname_single).replace('.', '_')

                msg = None

                try:
                    # for pyarrow
                    context = pa.default_serialization_context()

                    r = redis.StrictRedis(host=db_server, port=db_port, db=0)

                    # is there a compressed key stored?)
                    k = r.keys('comp_*_' + fname_single)

                    # if so, then it means that we have stored it as a compressed object
                    # if have more than 1 element, take the last (which will be the latest to be added)
                    if (len(k) >= 1):
                        k = k[-1].decode('utf-8')

                        comp = r.get(k)

                        siz = int(k.split('_')[1])
                        dec = pa.decompress(comp, codec='gzip', decompressed_size=siz)

                        # Now uses pyarrow
                        try:
                            context = pa
                            msg = context.deserialize(dec)

                        except:
                            # For earlier pyarrow versions
                            context = pa.default_serialization_context()
                            msg = context.deserialize(dec)
                    else:
                        msg = r.get(fname_single)

                        # print(fname_single)
                        if msg is not None:
                            try:
                                context = pa
                                msg = context.deserialize(msg)

                            except:
                                # For earlier pyarrow versions
                                context = pa.default_serialization_context()
                                msg = context.deserialize(msg)

                            # logger.warning("Key " + fname_single + " not in Redis cache?")

                except Exception as e:
                    logger.info("Cache not existent for " + fname_single + " in Redis: " + str(e))

                if msg is None:
                    data_frame = None
                else:
                    logger.info('Load Redis cache: ' + fname_single)

                    data_frame = msg # pd.read_msgpack(msg)

            elif (engine == 'arctic'):
                socketTimeoutMS = 2 * 1000

                import pymongo
                from arctic import Arctic

                fname_single = os.path.basename(fname_single).replace('.', '_')

                logger.info('Load Arctic/MongoDB library: ' + fname_single)

                if username is not None and password is not None:
                    c = pymongo.MongoClient(
                        host="mongodb://" + username + ":" + password + "@" + str(db_server) + ":" + str(db_port),
                        connect=False)  # , username=username, password=password)
                else:
                    c = pymongo.MongoClient(host="mongodb://" + str(db_server) + ":" + str(db_port), connect=False)

                store = Arctic(c, socketTimeoutMS=socketTimeoutMS, serverSelectionTimeoutMS=socketTimeoutMS)

                # Access the library
                try:
                    library = store[fname_single]

                    if start_date is None and finish_date is None:
                        item = library.read(fname_single)

                    else:
                        from arctic.date import DateRange
                        item = library.read(fname_single, date_range=DateRange(start_date.replace(tzinfo=None), finish_date.replace(tzinfo=None)))

                    c.close()

                    logger.info('Read ' + fname_single)

                    data_frame = item.data

                except Exception as e:
                    logger.warning('Library may not exist or another error: ' + fname_single + ' & message is ' + str(e))
                    data_frame = None

            elif self.path_exists(self.get_h5_filename(fname_single)):
                store = pd.HDFStore(self.get_h5_filename(fname_single))
                data_frame = store.select("data")

                if ('intraday' in fname_single):
                    data_frame = data_frame.astype('float32')

                store.close()

            elif self.path_exists(fname_single) and '.csv' in fname_single:
                data_frame = pd.read_csv(fname_single, index_col=0)

                data_frame.index = pd.to_datetime(data_frame.index)

            elif self.path_exists(fname_single):
                data_frame = self.read_parquet(fname_single)
                # data_frame = pd.read_parquet(fname_single)

            data_frame_list.append(data_frame)

        if len(data_frame_list) == 0:
            return None

        if len(data_frame_list) == 1:
            return data_frame_list[0]

        return data_frame_list

    ### functions for CSV reading and writing
    def write_time_series_to_csv(self, csv_path, data_frame):
        data_frame.to_csv(csv_path)

    def read_csv_data_frame(self, f_name, freq, cutoff=None, dateparse=None,
                            postfix='.close', intraday_tz='UTC', excel_sheet=None):
        """Reads CSV/Excel from disk into DataFrame

        Parameters
        ----------
        f_name : str
            CSV/Excel file path to read
        freq : str
            Frequency of data to read (intraday/daily etc)
        cutoff : DateTime (optional)
            end date to read up to
        dateparse : str (optional)
            date parser to use
        postfix : str (optional)
            postfix to add to each columns
        intraday_tz : str (optional)
            timezone of file if uses intraday data
        excel_sheet : str (optional)
            Excel sheet to be read

        Returns
        -------
        DataFrame
        """

        if (freq == 'intraday'):

            if dateparse is None:
                dateparse = lambda x: datetime.datetime(*map(int, [x[6:10], x[3:5], x[0:2],
                                                                   x[11:13], x[14:16], x[17:19]]))
            elif dateparse == 'dukascopy':
                dateparse = lambda x: datetime.datetime(*map(int, [x[0:4], x[5:7], x[8:10],
                                                                   x[11:13], x[14:16], x[17:19]]))
            elif dateparse == 'c':
                # use C library for parsing dates, several hundred times quicker
                # requires compilation of library to install
                import ciso8601
                dateparse = lambda x: ciso8601.parse_datetime(x)

            if excel_sheet is None:
                data_frame = pd.read_csv(f_name, index_col=0, parse_dates=True, date_parser=dateparse)
            else:
                data_frame = pd.read_excel(f_name, excel_sheet, index_col=0, na_values=['NA'])

            data_frame = data_frame.astype('float32')
            data_frame.index.names = ['Date']

            old_cols = data_frame.columns
            new_cols = []

            # add '.close' to each column name
            for col in old_cols:
                new_cols.append(col + postfix)

            data_frame.columns = new_cols
        else:
            # daily data
            if 'events' in f_name:

                data_frame = pd.read_csv(f_name)

                # very slow conversion
                data_frame = data_frame.convert_objects(convert_dates='coerce')

            else:
                if excel_sheet is None:
                    try:
                        data_frame = pd.read_csv(f_name, index_col=0, parse_dates=["DATE"], date_parser=dateparse)
                    except:
                        data_frame = pd.read_csv(f_name, index_col=0, parse_dates=["Date"], date_parser=dateparse)
                else:
                    data_frame = pd.read_excel(f_name, excel_sheet, index_col=0, na_values=['NA'])

        # convert Date to Python datetime
        # datetime data_frame['Date1'] = data_frame.index

        # slower method: lambda x: pd.datetime.strptime(x, '%d/%m/%Y %H:%M:%S')
        # data_frame['Date1'].apply(lambda x: datetime.datetime(int(x[6:10]), int(x[3:5]), int(x[0:2]),
        #                                        int(x[12:13]), int(x[15:16]), int(x[18:19])))

        # data_frame.index = data_frame['Date1']
        # data_frame.drop('Date1')

        # slower method: data_frame.index = pd.to_datetime(data_frame.index)

        if (freq == 'intraday'):
            # assume time series are already in UTC and assign this (can specify other time zones)
            data_frame = data_frame.tz_localize(intraday_tz)

        # end cutoff date
        if cutoff is not None:
            if (isinstance(cutoff, str)):
                cutoff = parse(cutoff)

            data_frame = data_frame.loc[data_frame.index < cutoff]

        return data_frame

    def find_replace_chars(self, array, to_find, replace_with):

        for i in range(0, len(to_find)):
            array = [x.replace(to_find[i], replace_with[i]) for x in array]

        return array

    def convert_csv_data_frame(self, f_name, category, freq, cutoff=None, dateparse=None):
        """Converts CSV file to HDF5 file

        Parameters
        ----------
        f_name : str
            File name to be read
        category : str
            data category of file (used in HDF5 filename)
        freq : str
            intraday/daily frequency (used in HDF5 filename)
        cutoff : DateTime (optional)
            filter dates up to here
        dateparse : str
            date parser to use
        """

        logger = LoggerManager().getLogger(__name__)

        logger.info("About to read... " + f_name)

        data_frame = self.read_csv_data_frame(f_name, freq, cutoff=cutoff, dateparse=dateparse)

        category_f_name = self.create_cache_file_name(category)

        self.write_time_series_cache_to_disk(category_f_name, data_frame)

    def clean_csv_file(self, f_name):
        """Cleans up CSV file (removing empty characters) before writing back to disk

        Parameters
        ----------
        f_name : str
            CSV file to be cleaned
        """
        logger = LoggerManager().getLogger(__name__)

        with codecs.open(f_name, 'rb', 'utf-8') as myfile:
            data = myfile.read()

            # clean file first if dirty
            if data.count('\x00'):
                logger.info('Cleaning CSV...')

                with codecs.open(f_name + '.tmp', 'w', 'utf-8') as of:
                    of.write(data.replace('\x00', ''))

                shutil.move(f_name + '.tmp', f_name)

    def create_cache_file_name(self, filename):
        return constants.folder_time_series_data + "/" + filename

    # TODO refactor IOEngine so that each database is implemented in a subclass of DBEngine

    def get_engine(self, engine='hdf5_fixed'):
        pass

    def sanitize_path(self, path):
        """Will remove unnecessary // from a file path (eg. in the middle)

        Parameters
        ----------
        path : str
            path to be sanitized

        Returns
        -------
        str
        """
        if "s3://" in path:
            path = path.replace("s3://", "")
            path = path.replace("//", "/")

            return "s3://" + path

        return path

    def read_parquet(self, path, cloud_credentials=None):
        """Reads a Pandas DataFrame from a local or s3 path

        Parameters
        ----------
        path : str
            Path of Parquet file (can be S3)

        cloud_credentials : dict (optional)
            Credentials for logging into the cloud

        Returns
        -------
        DataFrame
        """
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if "s3://" in path:
            storage_options = self._convert_cred(cloud_credentials, convert_to_s3fs=True)

            return pd.read_parquet(self.sanitize_path(path), storage_options=storage_options)
        else:
            return pd.read_parquet(path)


    def _create_cloud_filesystem(self, cloud_credentials, filesystem_type):

        cloud_credentials = self._convert_cred(cloud_credentials)

        # os.environ["AWS_ACCESS_KEY_ID"] = cloud_credentials['aws_access_key']
        # os.environ["AWS_SECRET_ACCESS_KEY"] = cloud_credentials['aws_secret_key']
        # os.environ["AWS_SESSION_TOKEN"] = cloud_credentials['aws_session_token']

        if 's3_pyarrow' == filesystem_type:
            return pyarrow.fs.S3FileSystem(anon=cloud_credentials['aws_anon'],
                                           access_key=cloud_credentials['aws_access_key'],
                                           secret_key=cloud_credentials['aws_secret_key'],
                                           session_token=cloud_credentials['aws_session_token'])

        elif 's3_filesystem' == filesystem_type:
            return S3FileSystem(anon=cloud_credentials['aws_anon'],
                                key=cloud_credentials['aws_access_key'],
                                secret=cloud_credentials['aws_secret_key'],
                                token=cloud_credentials['aws_session_token'])

    def _convert_cred(self, cloud_credentials, convert_to_s3fs=False):
        """Backfills the credential dictionary (usually for AWS login)
        """

        cloud_credentials = copy.copy(cloud_credentials)

        boolean_keys = {'aws_anon' : False}

        mappings = {'aws_anon': 'anon',
                    'aws_access_key': 'key',
                    'aws_secret_key': 'secret',
                    'aws_session_token': 'token'
                    }

        for m in mappings.keys():
            if m not in cloud_credentials.keys():
                if m in boolean_keys:
                    cloud_credentials[m] = boolean_keys[m]
                else:
                    cloud_credentials[m] = None

        # Converts the field names eg. aws_access_key => key etc.
        # Mainly for using pd.read_parquet
        if convert_to_s3fs:

            cloud_credentials_temp = {}

            for m in cloud_credentials.keys():
                cloud_credentials_temp[mappings[m]] = cloud_credentials[m]

            return cloud_credentials_temp

        return cloud_credentials

    def to_parquet(self, df, path, filename=None, cloud_credentials=None,
                   parquet_compression=constants.parquet_compression, use_pyarrow_directly=False):
        """Write a DataFrame to a local or s3 path as a Parquet file

        Parameters
        ----------
        df : DataFrame
            DataFrame to be written

        path : str(list)
            Paths where the DataFrame will be written

        filename : str (optional)
            Filename to be used (will be combined with the specified paths)

        cloud_credentials : str (optional)
            AWS credentials for S3 dump

        parquet_compression : str (optional)
            Parquet compression type to use when writing
        """
        logger = LoggerManager.getLogger(__name__)

        if cloud_credentials is None: cloud_credentials = DataConstants().cloud_credentials

        if isinstance(path, list):
            pass
        else:
            path = [path]

        if filename is not None:
            new_path = []

            for p in path:
                new_path.append(self.path_join(p, filename))

            path = new_path

        constants = DataConstants()

        # is_date = False
        #
        # # Force any date columns to default time units (Parquet with pyarrow has problems with ns dates)
        # for c in df.columns:
        #
        #     # If it's a date column don't append to convert to a float
        #     for d in constants.always_date_columns:
        #         if d in c or 'release-dt' in c:
        #             is_date = True
        #             break
        #
        #     if is_date:
        #         try:
        #             df[c] = pd.to_datetime(df[c], errors='coerce', unit=constants.default_time_units)
        #         except:
        #             pass

        try:
            df.index = pd.to_datetime(df.index, unit=constants.default_time_units)
        except:
            pass

        cloud_credentials_ = self._convert_cred(cloud_credentials)

        # Tends to be slower than using pandas/pyarrow directly, but for very large files, we might have to split
        # before writing to disk
        def pyarrow_dump(df, path):
            # Trying to convert large Pandas DataFrames in one go to Arrow tables can result in out-of-memory
            # messages, so chunk them first, convert them one by one, and write to disk in chunks
            df_list = self.chunk_dataframes(df)

            # Using pandas.to_parquet, doesn't let us pass in parameters to allow coersion of timestamps
            # hence have to do it this way, using underlying pyarrow interface (/)
            # ie. ns -> us
            if not (isinstance(df_list, list)):
                df_list = [df_list]

            for p in path:
                p = self.sanitize_path(p)

                # Reference:
                # https://stackoverflow.com/questions/47113813/using-pyarrow-how-do-you-append-to-parquet-file
                # https://arrow.apache.org/docs/python/filesystems.html

                pqwriter = None
                counter = 1

                if 's3://' in p:
                    s3 = self._create_cloud_filesystem(cloud_credentials_, 's3_pyarrow')

                    p_in_s3 = p.replace("s3://", "")

                    for df_ in df_list:
                        logger.info("S3 chunk... " + str(counter) + " of " + str(len(df_list)))
                        table = pa.Table.from_pandas(df_)

                        if pqwriter is None:
                            pqwriter = pq.ParquetWriter(p_in_s3, table.schema, compression=parquet_compression,
                                                        coerce_timestamps=constants.default_time_units,
                                                        allow_truncated_timestamps=True, filesystem=s3)

                        pqwriter.write_table(table)

                        counter = counter + 1

                else:
                    for df_ in df_list:
                        logger.info("Local chunk... " + str(counter) + " of " + str(len(df_list)))
                        table = pa.Table.from_pandas(df_)

                        if pqwriter is None:
                            pqwriter = pq.ParquetWriter(p, table.schema, compression=parquet_compression,
                                                        coerce_timestamps=constants.default_time_units,
                                                        allow_truncated_timestamps=True)

                        pqwriter.write_table(table)

                        counter = counter + 1

                # Close the parquet writer
                if pqwriter:
                    pqwriter.close()

                    # df.to_parquet(path, compression=parquet_compression)

        if use_pyarrow_directly:
            pyarrow_dump(df, path)
        else:
            # First try to use Pandas/pyarrow, if fails, which can occur with large DataFrames use chunked write
            try:
                for p in path:
                    p = self.sanitize_path(p)

                    if "s3://" in p:
                        storage_options = self._convert_cred(cloud_credentials, convert_to_s3fs=True)

                        df.to_parquet(p, compression=parquet_compression,
                                      coerce_timestamps=constants.default_time_units,
                                      allow_truncated_timestamps=True, storage_options=storage_options)
                    else:

                        df.to_parquet(p, compression=parquet_compression,
                                          coerce_timestamps=constants.default_time_units,
                                          allow_truncated_timestamps=True)

            except pyarrow.lib.ArrowMemoryError as e:
                logger.warning("Couldn't dump using Pandas/pyarrow, will instead try chunking with pyarrow directly " + str(e))

                pyarrow_dump(df, path)


    def split_array_chunks(self, array, chunks=None, chunk_size=None):
        """Splits an array or DataFrame into a list of equally sized chunks

        Parameters
        ----------
        array : NumPy array/pd DataFrame
            array to be split into chunks

        chunks : int (optional)
            number of chunks

        chunk_size : int (optional)
            size of each chunk (in rows)

        Returns
        -------
        list of arrays or DataFrames
        """

        if chunk_size is None and chunks is None:
            return array

        if chunk_size is None:
            chunk_size = int(array.shape[0] / chunks)

        if chunks is None:
            chunks = int(array.shape[0] / chunk_size)

        # alternative split array method (untested)

        # if isinstance(array, pd.DataFrame):
        #     array = array.copy()
        #     array_list = []
        #
        #     for start in range(0, array.shape[0], chunk_size):
        #         array_list.append(array.iloc[start:start + chunk_size])
        #
        #     return array_list

        if chunks > 0:
            # if isinstance(array, pd.DataFrame):
            #    array = [array[i:i + chunk_size] for i in range(0, array.shape[0], chunk_size)]

            return np.array_split(array, chunks)

        return array

    def get_obj_size_mb(self, obj):
        # Can sometime have very large dataframes, which need to be split, otherwise won't fit in a single Redis key
        mem = obj.memory_usage(deep='deep').sum()
        mem_float = round(float(mem) / (1024.0 * 1024.0), 3)

        return mem_float

    def chunk_dataframes(self, obj, chunk_size_mb=constants.chunk_size_mb):
        logger = LoggerManager.getLogger(__name__)

        # Can sometime have very large dataframes, which need to be split, otherwise won't fit in a single Redis key
        mem_float = self.get_obj_size_mb(obj)
        mem = '----------- ' + str(mem_float) + ' MB -----------'

        chunks = int(math.ceil(mem_float / chunk_size_mb))

        if chunks > 1:
            obj_list = self.split_array_chunks(obj, chunks=chunks)
        else:

            obj_list = [obj]

        if obj_list != []:
            logger.info("Pandas dataframe of size: " + mem + " in " + str(chunks) + " chunk(s)")

        return obj_list

    def read_csv(self, path, cloud_credentials=None, encoding='utf-8', encoding_errors=None, errors='ignore'):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if "s3://" in path:
            s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

            path_in_s3 = self.sanitize_path(path).replace("s3://", "")

            # Use 'w' for py3, 'wb' for py2
            with s3.open(path_in_s3, 'r', errors=errors) as f:
                if encoding_errors is not None:
                    return pd.read_csv(f, encoding=encoding, encoding_errors=encoding_errors)
                else:
                    return pd.read_csv(f, encoding=encoding)
        else:
            if encoding_errors is not None:
                return pd.read_csv(path, encoding=encoding, encoding_errors=encoding_errors)
            else:
                return pd.read_csv(path, encoding=encoding)

    def to_csv(self, df, path, filename=None, cloud_credentials=None):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if isinstance(path, list):
            pass
        else:
            path = [path]

        if filename is not None:
            new_path = []

            for p in path:
                new_path.append(self.path_join(p, filename))

            path = new_path

        for p in path:
            if "s3://" in p:
                s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

                path_in_s3 = self.sanitize_path(p).replace("s3://", "")

                # Use 'w' for py3, 'wb' for py2
                with s3.open(path_in_s3, 'w') as f:
                    df.to_csv(f)
            else:
                df.to_csv(p)

    def path_exists(self, path, cloud_credentials=None):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if "s3://" in path:
            s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

            path_in_s3 = path.replace("s3://", "")

            return s3.exists(path_in_s3)
        else:
            return os.path.exists(path)

    def path_join(self, folder, file):
        if 's3://' in folder:
            folder = folder.replace("s3://", "")
            folder = folder + "/" + file

            folder = folder.replace("//", "/")

            folder = "s3://" + folder

        else:
            if file[0] == '/':
                file = file[1::]

            folder = os.path.join(folder, file)

        folder = folder.replace("\\\\", "/")
        folder = folder.replace("\\", "/")

        return folder

    def list_files(self, path, cloud_credentials=None):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if "s3://" in path:
            s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

            path_in_s3 = self.sanitize_path(path).replace("s3://", "")

            list_files = s3.glob(path_in_s3)

            if path_in_s3 in list_files:
                list_files.remove(path_in_s3)

            files = ['s3://' + x for x in list_files]

        else:
            files = glob.glob(path)

        list.sort(files)

        return files

    def delete(self, path, cloud_credentials=None):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if not(isinstance(path, list)):
            path = [path]

        for p in path:
            if "s3://" in p:
                s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

                path_in_s3 = self.sanitize_path(p).replace("s3://", "")

                if self.path_exists(path, cloud_credentials=cloud_credentials):
                    s3.delete(path_in_s3)
            else:
                if self.path_exists(p, cloud_credentials=cloud_credentials):
                    os.remove(p)

    def copy(self, source, destination, cloud_credentials=None, infer_dest_filename=False):
        if cloud_credentials is None: cloud_credentials = constants.cloud_credentials

        if destination is None:
            destination = ""
            infer_dest_filename = True

        if not(isinstance(source, list)):
            source = [source]

        for so in source:
            dest = destination

            # Special case for wildcard *
            if "*" in so:
                # Infer filename if destination in s3
                if "s3://" in dest:
                    infer_dest_filename = True

                list_files = self.list_files(so, cloud_credentials=cloud_credentials)

                self.copy(list_files, dest, infer_dest_filename=infer_dest_filename)
            else:
                if infer_dest_filename:
                    dest = self.path_join(destination, os.path.basename(so))

                if "s3://" not in dest and "s3://" not in so:
                    shutil.copy(so, dest)
                else:
                    s3 = self._create_cloud_filesystem(cloud_credentials, 's3_filesystem')

                    if "s3://" in dest and "s3://" in so:
                        s3.cp(self.sanitize_path(so).replace("s3://", ""),
                              self.sanitize_path(dest).replace("s3://", ""), recursive=True)

                    elif "s3://" in dest and "s3://" not in so:
                        s3.put(so, self.sanitize_path(dest).replace("s3://", ""), recursive=True)

                    elif "s3://" not in dest and "s3://" in so:
                        s3.get(self.sanitize_path(so).replace("s3://", ""), dest, recursive=True)


#######################################################################################################################

class SpeedCache(object):
    """Wrapper for cache hosted in external in memory database (by default Redis, although in practice, can use
    any database supported in this class). This allows us to share hash across Python instances, rather than having
    repopulate each time we restart Python. Also can let us share cache easily across threads, without replicating.

    """

    def __init__(self, db_cache_server=None, db_cache_port=None, engine='redis'):

        if db_cache_server is None:
            self.db_cache_server = constants.db_cache_server

        if db_cache_port is None:
            self.db_cache_port = constants.db_cache_port

        self.engine = engine
        self.io_engine = IOEngine()

    def put_dataframe(self, key, obj):
        if self.engine != 'no_cache':
            try:
                self.io_engine.write_time_series_cache_to_disk(key.replace('/', '_'), obj,
                                                               engine=self.engine, db_server=self.db_cache_server,
                                                               db_port=self.db_cache_port)
            except:
                pass

    def get_dataframe(self, key):
        if self.engine == 'no_cache': return None

        try:
            return self.io_engine.read_time_series_cache_from_disk(key.replace('/', '_'),
                                                                   engine=self.engine, db_server=self.db_cache_server,
                                                                   db_port=self.db_cache_port)
        except:
            pass

    def dump_all_keys(self):
        self.dump_key('flush_all_keys')

    def dump_key(self, key):
        if self.engine == 'no_cache': return

        try:
            return self.io_engine.remove_time_series_cache_on_disk(key, engine=self.engine,
                                                                   db_server=self.db_cache_server,
                                                                   db_port=self.db_cache_port)
        except:
            pass

    def generate_key(self, obj, key_drop=[]):
        """Create a unique hash key for object from its attributes (excluding those attributes in key drop), which can be
        used as a hashkey in the Redis hashtable

        Parameters
        ----------
        obj : class
            Any Python class
        key_drop : str (list)
            List of internal attributes to drop before hashing

        Returns
        -------
        hashkey
        """

        # never want to include Logger object!
        key_drop.append('logger')
        key = []

        for k in obj.__dict__:

            if 'api_key' not in k:
                # provided the key is not in one of the dropped keys
                if not(any(a == k for a in key_drop)):
                    add = obj.__dict__[k]

                    if add is not None:
                        if isinstance(add, list): add = '_'.join(str(e) for e in add)

                    key.append(str(k) + '-' + str(add))

        key.sort()
        key = '_'.join(str(e) for e in key).replace(type(obj).__name__, '').replace('___', '_')

        return type(obj).__name__ + "_" + str(len(str(key))) + "_" + str(key)


# TODO refactor code to use DBEngine
class DBEngine(object):
    pass


class DBEngineArctic(DBEngine):
    pass


class DBEngineHDF5(DBEngine):
    pass


class DBRedis(DBEngine):
    pass



import pytest
import pandas as pd

from findatapy.market import Market, MarketDataGenerator, MarketDataRequest
from findatapy.util.dataconstants import DataConstants

market = Market(market_data_generator=MarketDataGenerator())

data_constants = DataConstants()
quandl_api_key = data_constants.quandl_api_key

def test_quandl_download():
    md_request = MarketDataRequest(start_date='year', category='fx', data_source='quandl', tickers=['AUDJPY'],
                                   quandl_api_key=quandl_api_key)

    df = market.fetch_market(md_request)

    assert df is not None

def test_yahoo_download():
    md_request = MarketDataRequest(
        start_date="year",  # start date
        data_source='yahoo',  # use Bloomberg as data source
        tickers=['Apple', 'Citigroup'],  # ticker (findatapy)
        fields=['close'],  # which fields to download
        vendor_tickers=['aapl', 'c'],  # ticker (Yahoo)
        vendor_fields=['Close'])  # which Bloomberg fields to download)

    df = market.fetch_market(md_request)

    assert df is not None

def test_free_fx_tick_download():
    # first we can do it by defining all the vendor fields, tickers etc. so we bypass the configuration file
    md_request = MarketDataRequest(start_date='05 Dec 2016', finish_date='07 Dec 2016',
                                   fields=['bid'], vendor_fields=['bid'],
                                   freq='tick', data_source='fxcm',
                                   tickers=['EURUSD'], vendor_tickers=['EURUSD'])

    df_tick = market.fetch_market(md_request)
    df = df_tick.resample('1min').mean()

    md_request.data_source = 'dukascopy'
    df1 = market.fetch_market(md_request).resample('1min').mean()

    assert df is not None
    assert df1 is not None

if __name__ == '__main__':
    pytest.main()

    # test_quandl_download()
    # test_yahoo_download()
    # test_free_fx_tick_download()




import pytest
import pandas as pd

from findatapy.util.dataconstants import DataConstants

data_constants = DataConstants()

redis_server = data_constants.db_cache_server
redis_port = data_constants.db_cache_port

def test_redis_caching():
    # Note: you need to install Redis in order for this to work!

    # read CSV from disk, and make sure to parse dates
    df = pd.read_csv("S&P500.csv", parse_dates=['Date'], index_col=['Date'])
    df.index = pd.to_datetime(df.index)

    from findatapy.market.ioengine import IOEngine

    io = IOEngine()

    use_cache_compression = [True, False]

    for u in use_cache_compression:
        # Write DataFrame to Redis (using pyarrow format)
        io.write_time_series_cache_to_disk('test_key', df, engine='redis', db_server=redis_server, db_port=redis_port,
                                           use_cache_compression=u)

        # Read back DataFrame from Redis (using pyarrow format)
        df_out = io.read_time_series_cache_from_disk('test_key', engine='redis', db_server=redis_server, db_port=redis_port)

        pd.testing.assert_frame_equal(df, df_out)

if __name__ == '__main__':
    # pytest.main()

    test_redis_caching()


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool;

    SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download event data for US payrolls
    # run_example = 2 - download different GDP releases (advance, preliminary and final) for US
    run_example = 2

    if run_example == 1 or run_example == 0:
        # Download event data from Bloomberg
        # we have to use the special category "events" keyword for economic data events
        # so findatapy can correctly identify them (given the underlying Bloomberg API calls are all different, however,
        # this will appear transparent to the user)
        market_data_request = MarketDataRequest(
            start_date="year",
            category="events",
            data_source='bloomberg',  # Use Bloomberg as data source
            tickers=['FOMC', 'NFP'],
            fields=['release-date-time-full', 'release-dt', 'actual-release'],  # which fields to download
            vendor_tickers=['FDTR Index', 'NFP TCH Index'],  # ticker (Bloomberg)
            vendor_fields=['ECO_FUTURE_RELEASE_DATE_LIST', 'ECO_RELEASE_DT',
                           'ACTUAL_RELEASE'])  # which Bloomberg fields to download

        df = market.fetch_market(market_data_request)

        print(df)

        # We also have a few events defined in our configuation file
        # those tickers/fields which are predefined this way are easier to download
        # note how we don't have to use the vendor_tickers and vendor_fields for examples
        market_data_request = MarketDataRequest(
            start_date="year",
            category="events",
            data_source='bloomberg',  # use Bloomberg as data source
            tickers=['USD-US Employees on Nonfarm Payrolls Total MoM Net Change SA'],
            fields=['release-date-time-full', 'release-dt', 'actual-release', 'number-observations'])

        df = market.fetch_market(market_data_request)

        print(df)

        # Now just download the event day
        market_data_request = MarketDataRequest(
            start_date="year",
            category="events",
            data_source='bloomberg',  # use Bloomberg as data source
            tickers=['NFP'],
            fields=['release-date-time-full'],  # which fields to download
            vendor_tickers=['NFP TCH Index'],  # ticker (Bloomberg)
            vendor_fields=['ECO_FUTURE_RELEASE_DATE_LIST'])  # which Bloomberg fields to download

        df = market.fetch_market(market_data_request)

        print(df)

    if run_example == 2 or run_example == 0:
        # Download Advance, Preliminary and Final estimates for US GDP and US payrolls (
        # With Bloomberg, we internally set the correct overrides to fetch the correct series for GDP
        market_data_request = MarketDataRequest(
            start_date="01 Jan 2007",
            data_source='bloomberg',  # use Bloomberg as data source
            tickers=['US GDP Advance QoQ', 'US GDP Preliminary QoQ', 'US GDP Final QoQ', 'US payrolls'],
            fields=['actual-release', 'survey-median'],  # which fields to download
            vendor_tickers=['GDP CQOQ Index', 'GDP CQOQ Index', 'GDP CQOQ Index', 'NFP TCH Index'],  # ticker (Bloomberg)
            vendor_fields=['ACTUAL_RELEASE', 'BN_SURVEY_MEDIAN'])  # which Bloomberg fields to download

        df_gdp_vintages = market.fetch_market(market_data_request)

        print(df_gdp_vintages)



__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2021 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # We can grab market data, by using MarketDataRequest objects
    # However, sometimes we might wish to create more freeform requests
    #   eg. if we have tickers stored in a spreadsheet)
    #   eg. from a single string (either for predefined tickers or any tickers/vendor_tickers combinations we want)
    # Here, we show how to this, and also how to call tickers which are already predefined in the CSV config files

    # Now we will try various examples with Bloomberg
    # only works if you have Bloomberg terminal installed and the Python API!

    # In this case, we are using the special 'fx' category
    md_request = MarketDataRequest(start_date='week', category='fx', data_source='bloomberg')

    # Let's try creating a MarketDataRequest overwrites in the tickers field
    md_request.freeform_md_request = [{'tickers' : ['AUDJPY']}, {'tickers' : ['AUDJPY', 'AUDUSD']}]
    md_request_list = market.create_md_request_from_freeform(md_request)

    print(md_request_list)

    # Now, we can do the same thing with a more complicated overwrite
    md_request.freeform_md_request = [{'tickers' : ['AUDJPY'], 'fields' : 'close'},
                                      {'tickers' : ['AUDJPY', 'AUDUSD'], 'fields' : 'close'}]

    md_request_list = market.create_md_request_from_freeform(md_request)

    print(md_request_list)

    # We can also create MarketDataRequest by passing a str that relates to predefined tickers
    md_request = market.create_md_request_from_str('backtest.fx.bloomberg.daily.NYC.EURUSD.close')

    print(md_request)

    # We can do an approximate match for predefined tickers, and findatapy will "guess" the closest match
    md_request = market.create_md_request_from_str('_.bloomberg.EURUSD.NYC')

    print(md_request)

    # We get *all* the predefined tickers which match in any column with quandl and fx, and these be smart grouped
    # into the smallest number of requests
    md_request = market.create_md_request_from_str('_.quandl.fx', best_match_only=False, smart_group=True)

    print(md_request)

    # We can also create MarketDataRequest by passing a str for an arbitrary Parquet, note, if we have dots in the path
    # we need to use {} to denote them (we can use either 'raw' to denote this or just 'r')
    md_request = market.create_md_request_from_str('raw.data_source.{c:\parquet_files\dump.parquet}.tickers.EURUSD')

    print(md_request)

    # We can also create a MarketDataRequest using a JSON string (we must be careful to adhere to correct format)
    # where the keys match those available of a MarketDataRequest
    md_request = market.create_md_request_from_str('''{"data_source" : "bloomberg",
        "category" : "fx",
        "freq" : "daily",
        "tickers" : ["EURUSD", "GBPUSD"],
        "cut" : "NYC",
        "fields" : "close"}''')

    print(md_request)

    # Or directly from a dict where the keys match those available of a MarketDataRequest (again we need to be careful
    # with the keys, so they match those of a MarketDataRequest
    md_request = market.create_md_request_from_dict({
        "data_source" : "bloomberg",
        "category" : "fx",
        "freq" : "daily",
        "tickers" : ["EURUSD", "GBPUSD"],
        "cut" : "NYC",
        "fields" : "close"})

    print(md_request)

    # Or we can do a raw request using any properties we want. At a minimum, we need to have the data_source, tickers,
    # and vendor_tickers field, any ones we don't include, will just be taken from the defaults
    md_request = market.create_md_request_from_str('raw.data_source.bloomberg.tickers.EURUSD.vendor_tickers.EURUSD Curncy')

    print(md_request)

    # Or we can directly call fetch_market with it
    # Format of the call is environment.category.freq.cut.ticker.field which is for predefined tickers

    # Note, we can choose to omit the environment if we want and findatapy will choose the default environment
    # which is backtest (in this case, we have included it)
    print(market.fetch_market('backtest.fx.bloomberg.daily.NYC.EURUSD.close'))
    print(market.fetch_market('backtest.fx.quandl.daily.NYC.EURUSD.close'))

    # We can also pass in a MarketDataRequest which can contain many more parameters, which we can't specify
    # with a string of the above form. Given we've ignored adding the environment here, findatapy will automatically
    # add the default environment which is backtest
    print(market.fetch_market(md_request_str='fx.bloomberg.intraday.NYC.EURUSD',
                              md_request=MarketDataRequest(fields=['open', 'high', 'low', 'close'])))

    # Let's do that for whole vol surface for EURUSD (note, we have omitted backtest here, which findatapy assumes as default)
    print(market.fetch_market('fx-implied-vol.bloomberg.daily.BGN.EURUSD.close',
                              start_date='30 Apr 2021', finish_date='30 Apr 2021'))

    # We can also download any arbitary properties such as tickers/vendor tickers which aren't predefined
    print(market.fetch_market('raw.data_source.bloomberg.tickers.VIX.vendor_tickers.VIX Index',
                              start_date='01 Apr 2021', finish_date='30 Apr 2021'))

    # Or let's give it a Python dict like above, which is likely to be easier to structure, when we have many properties
    print(market.fetch_market({
        "start_date" : "01 Apr 2021",
        "finish_date" : "30 Apr 2021",
        "category": "fx",
        "data_source" : "bloomberg",
        "freq" : "daily",
        "tickers" : ["EURUSD", "GBPUSD"],
        "cut" : "NYC",
        "fields" : "close"}))




from findatapy.market import MarketDataGenerator, Market, MarketDataRequest

def generate_market_data_for_tests():

    # generate daily S&P500 data from Quandl
    md_request = MarketDataRequest(start_date='01 Jan 2001', finish_date='01 Dec 2008',
                                   tickers=['S&P500'], vendor_tickers=['YAHOO/INDEX_GSPC'], fields=['close'],
                                   data_source='quandl')

    market = Market(market_data_generator=MarketDataGenerator())

    df = market.fetch_market(md_request)
    df.to_csv("S&P500.csv")

    # generate tick data from DukasCopy for EURUSD
    md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='15 Jun 2016', cut='NYC', category='fx',
                                   fields=['bid'], freq='tick', data_source='dukascopy',
                                   tickers=['EURUSD'])

    market = Market(market_data_generator=MarketDataGenerator())

    df = market.fetch_market(md_request)
    df.to_csv("EURUSD_tick.csv")

if __name__ == '__main__':
    generate_market_data_for_tests()


'''

We can set API keys by running this script, these will be read in by dataconstants.py

Alternatively, we can set them manually in dataconstants.py or create a datacred.py file

'''

import os
import keyring

service_names = ['Quandl', 'AlphaVantage', 'Twitter App Key', 'Twitter App Secret', 'Twitter OAUTH token',
                 'Twitter OAUTH token Secret', 'FRED']

for s in service_names:
    key = input("Please enter the %s API key: " % s)

    keyring.set_password(s, os.getlogin(), key)


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download minute crypto data from Bloomberg
    # coming soon other crypto data sources!

    run_example = 1

    if run_example == 1 or run_example == 0:
        ####### Bloomberg

        # now we will try various examples with Bloomberg
        # only works if you have Bloomberg terminal installed and the Python API!
        md_request = MarketDataRequest(start_date='day', freq='intraday', category='fx', data_source='bloomberg', cut='BSTP',
                                       tickers=['XBTUSD', 'XETUSD', 'XLCUSD', 'XRPUSD'])

        df = market.fetch_market(md_request)

        print(df.tail(n=10))


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import math
import pandas
import collections

from findatapy.timeseries.calculations import Calculations


class RetStats(object):
    """Calculating return statistics of a time series

    """

    def __init__(self, returns_df=None, ann_factor=None, resample_freq=None):
        self._returns_df = returns_df
        self._ann_factor = ann_factor
        self._resample_freq = resample_freq

        self._rets = None
        self._vol = None
        self._inforatio = None
        self._kurtosis = None
        self._dd = None
        self._yoy_rets = None

    def split_into_dict(self):
        """If we have multiple columns in our returns, we can opt to split up the RetStats object into a dictionary of
        smaller RetStats object, one for each asset return

        Returns
        -------
        dict
            Dictionary of RetStats objects
        """

        ret_stats_dict = collections.OrderedDict()

        for d in self._returns_df.columns:

            returns_df = pandas.DataFrame(self._returns_df[d])

            # if column is of the form asset / signal, just keep the asset part
            try:
                d = d.split(' / ')[0]
            except:
                pass

            returns_df.columns = [d]

            ret_stats_dict[d] = RetStats(returns_df, self._ann_factor)

        return ret_stats_dict

    def calculate_ret_stats_from_prices(self, prices_df, ann_factor):
        """Calculates return statistics for an asset's price

        Parameters
        ----------
        prices_df : DataFrame
            asset prices
        ann_factor : int
            annualisation factor to use on return statistics

        Returns
        -------
        DataFrame
        """
        calculations = Calculations()

        self.calculate_ret_stats(calculations.calculate_returns(prices_df), ann_factor)

    def calculate_ret_stats(self, returns_df=None, ann_factor=None):
        """Calculates return statistics for an asset's returns including IR, vol, ret and drawdowns

        Parameters
        ----------
        returns_df : DataFrame
            asset returns
        ann_factor : int
            annualisation factor to use on return statistics

        Returns
        -------
        DataFrame
        """

        if returns_df is None: returns_df = self._returns_df
        if ann_factor is None: ann_factor = self._ann_factor

        if self._resample_freq is not None:
            returns_df = returns_df.resample(self._resample_freq).sum()

        # TODO work on optimizing this method
        self._rets = returns_df.mean(axis=0) * ann_factor
        self._vol = returns_df.std(axis=0) * math.sqrt(ann_factor)
        self._inforatio = self._rets / self._vol
        self._kurtosis = returns_df.kurtosis(axis=0) / math.sqrt(ann_factor)

        index_df = (1.0 + returns_df).cumprod()

        if pandas.__version__ < '0.17':
            max2here = pandas.expanding_max(index_df)
        else:
            max2here = index_df.expanding(min_periods=1).max()

        dd2here = index_df / max2here - 1

        self._dd = dd2here.min()
        self._yoy_rets = index_df.resample('Y').last().pct_change()

        return self

    def ann_returns(self):
        """Gets annualised returns

        Returns
        -------
        float
        """

        if self._rets is None: self.calculate_ret_stats()

        return self._rets

    def ann_vol(self):
        """Gets annualised volatility

        Returns
        -------
        float
        """
        if self._vol is None: self.calculate_ret_stats()

        return self._vol

    def inforatio(self):
        """Gets information ratio

        Returns
        -------
        float
        """

        if self._inforatio is None: self.calculate_ret_stats()

        return self._inforatio

    def drawdowns(self):
        """Gets drawdowns for an asset or strategy

        Returns
        -------
        float
        """
        if self._dd is None: self.calculate_ret_stats()

        return self._dd

    def kurtosis(self):
        """Gets kurtosis for an asset or strategy

        Returns
        -------
        float
        """
        if self._kurtosis is None: self.calculate_ret_stats()

        return self._kurtosis

    def yoy_rets(self):
        """Calculates the yoy rets

        Returns
        -------
        float
        """
        if self._yoy_rets is None: self.calculate_ret_stats()

        return self._yoy_rets

    def summary(self):
        """Gets summary string contains various return statistics

        Returns
        -------
        str
        """

        if self._rets is None: self.calculate_ret_stats()

        stat_list = []

        for i in range(0, len(self._rets.index)):
            stat_list.append(self._rets.index[i] + " Ret = " + str(round(self._rets[i] * 100, 1))
                             + "% Vol = " + str(round(self._vol[i] * 100, 1))
                             + "% IR = " + str(round(self._inforatio[i], 2))
                             + " Dr = " + str(round(self._dd[i] * 100, 1))
                             + "%") # Kurt = " + str(round(self._kurtosis[i], 2)))

        return stat_list



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import abc
import copy

from findatapy.market.marketdatarequest import MarketDataRequest
from findatapy.util import ConfigManager, LoggerManager

class DataVendor(object):
    """Abstract class for various data source loaders.

    """
    def __init__(self):
        self.config = ConfigManager().get_instance()
        # self.config = None
        return

    @abc.abstractmethod
    def load_ticker(self, market_data_request):
        """Retrieves market data from external data source

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        DataFrame
        """

        return

        # to be implemented by subclasses

    @abc.abstractmethod
    def kill_session(self):
        return

    def construct_vendor_market_data_request(self, market_data_request):
        """Creates a MarketDataRequest with the vendor tickers

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        MarketDataRequest
        """

        symbols_vendor = self.translate_to_vendor_ticker(market_data_request)
        fields_vendor = self.translate_to_vendor_field(market_data_request)

        market_data_request_vendor = MarketDataRequest(md_request=market_data_request)

        market_data_request_vendor.tickers = symbols_vendor
        market_data_request_vendor.fields = fields_vendor

        market_data_request_vendor.old_tickers = market_data_request.tickers

        return market_data_request_vendor

    def translate_to_vendor_field(self, market_data_request):
        """Converts all the fields from findatapy fields to vendor fields

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        List of Strings
        """

        if market_data_request.vendor_fields is not None:
            return market_data_request.vendor_fields

        source = market_data_request.data_source
        fields_list = market_data_request.fields

        if isinstance(fields_list, str):
            fields_list = [fields_list]

        if self.config is None: return fields_list

        fields_converted = []

        for field in fields_list:
            try:
                f = self.config.convert_library_to_vendor_field(source, field)
            except:
                logger = LoggerManager().getLogger(__name__)
                logger.warn("Couldn't find field conversion, did you type it correctly: " + field)

                return

            fields_converted.append(f)

        return fields_converted

    # Translate findatapy ticker to vendor ticker
    def translate_to_vendor_ticker(self, market_data_request):
        """Converts all the tickers from findatapy tickers to vendor tickers

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        List of Strings
        """

        if market_data_request.vendor_tickers is not None:
            return market_data_request.vendor_tickers

        category = market_data_request.category
        source = market_data_request.data_source
        freq = market_data_request.freq
        cut = market_data_request.cut
        tickers_list = market_data_request.tickers

        if isinstance(tickers_list, str):
            tickers_list = [tickers_list]

        if self.config is None: return tickers_list

        tickers_list_converted = []

        for ticker in tickers_list:
            try:
                t = self.config.convert_library_to_vendor_ticker(category, source, freq, cut, ticker)
            except:
                logger = LoggerManager().getLogger(__name__)
                logger.error("Couldn't find ticker conversion, did you type it correctly: " + ticker)

                return

            tickers_list_converted.append(t)

        return tickers_list_converted

    def translate_from_vendor_field(self, vendor_fields_list, market_data_request):
        """Converts all the fields from vendors fields to findatapy fields

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        List of Strings
        """

        data_source = market_data_request.data_source

        if isinstance(vendor_fields_list, str):
            vendor_fields_list = [vendor_fields_list]

        # if self.config is None: return vendor_fields_list

        fields_converted = []

        # If we haven't set the configuration files for automatic configuration
        if market_data_request.vendor_fields is not None:

            dictionary = dict(zip(self.get_lower_case_list(market_data_request.vendor_fields), market_data_request.fields))

            for vendor_field in vendor_fields_list:
                try:
                    fields_converted.append(dictionary[vendor_field.lower()])
                except:
                    fields_converted.append(vendor_field)

        # Otherwise used stored configuration files (every field needs to be defined!)
        else:
            for vendor_field in vendor_fields_list:
                try:
                    v = self.config.convert_vendor_to_library_field(data_source, vendor_field)
                except:
                    logger = LoggerManager().getLogger(__name__)
                    logger.error("Couldn't find field conversion, did you type it correctly: " + vendor_field + ", using 'close' as default.")

                    v = 'close'

                fields_converted.append(v)

        return fields_converted

    # Translate findatapy ticker to vendor ticker
    def translate_from_vendor_ticker(self, vendor_tickers_list, md_request):
        """Converts all the fields from vendor tickers to findatapy tickers

        Parameters
        ----------
        md_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        List of Strings
        """

        if md_request.vendor_tickers is not None:

            dictionary = dict(zip(self.get_lower_case_list(md_request.vendor_tickers), md_request.tickers))

            tickers_stuff = []

            for vendor_ticker in vendor_tickers_list:
                tickers_stuff.append(dictionary[vendor_ticker.lower()])

            return tickers_stuff # [item for sublist in tickers_stuff for item in sublist]


        # tickers_list = md_request.tickers

        if isinstance(vendor_tickers_list, str):
            vendor_tickers_list = [vendor_tickers_list]

        if self.config is None: return vendor_tickers_list

        tickers_converted = []

        for vendor_ticker in vendor_tickers_list:
            try:
                v = self.config.convert_vendor_to_library_ticker(
                    md_request.category, md_request.data_source,
                    md_request.freq, md_request.cut, vendor_ticker)
            except:
                logger = LoggerManager().getLogger(__name__)
                logger.error("Couldn't find ticker conversion, did you type it correctly: " + vendor_ticker)

                return

            tickers_converted.append(v)

        return tickers_converted

    def get_lower_case_list(self, lst):
        return [k.lower() for k in lst]


